from wfuzz.externals.moduleman.plugin import moduleman_plugin
from wfuzz.exception import FuzzExceptPluginBadParams, FuzzExceptBadInstall
from wfuzz.plugin_api.base import BasePayload
from wfuzz.fuzzobjects import FuzzWordType


@moduleman_plugin
class iprange(BasePayload):
    name = "iprange"
    author = ("Xavi Mendez (@xmendez)",)
    version = "0.1"
    description = (
        "ie. 192.168.1.0-192.168.1.12",
        "Requires: netaddr module",
    )
    summary = "Returns list of IP addresses of a given IP range."
    category = ["default"]
    priority = 99

    parameters = (
        ("iprange", "", True, "IP address range int the form 192.168.1.0-192.168.1.12"),
    )

    default_parameter = "iprange"

    def __init__(self, params):
        BasePayload.__init__(self, params)

        try:
            from netaddr import IPRange
            from netaddr.core import AddrFormatError

            ran = self.params["iprange"].split("-")
            net = IPRange(ran[0], ran[1])
            self.f = iter(net)
            self.__count = net.size
        except ImportError:
            raise FuzzExceptBadInstall(
                "ipnet plugin requires netaddr module. Please install it using pip."
            )
        except AddrFormatError:
            raise FuzzExceptPluginBadParams(
                "The specified network range has an incorrect format."
            )
        except IndexError:
            raise FuzzExceptPluginBadParams(
                "The specified network range has an incorrect format."
            )

    def get_type(self):
        return FuzzWordType.WORD

    def get_next(self):
        return str(next(self.f))

    def count(self):
        return self.__count



from wfuzz.plugin_api.mixins import DiscoveryPluginMixin
from wfuzz.plugin_api.base import BasePlugin
from wfuzz.exception import FuzzExceptResourceParseError
from wfuzz.externals.moduleman.plugin import moduleman_plugin

import xml.dom.minidom


@moduleman_plugin
class sitemap(BasePlugin, DiscoveryPluginMixin):
    name = "sitemap"
    author = ("Xavi Mendez (@xmendez)",)
    version = "0.1"
    summary = "Parses sitemap.xml file"
    description = ("Parses sitemap.xml file",)
    category = ["active", "discovery"]
    priority = 99

    parameters = ()

    def __init__(self):
        BasePlugin.__init__(self)

    def validate(self, fuzzresult):
        return (
            fuzzresult.history.urlparse.ffname == "sitemap.xml"
            and fuzzresult.code == 200
        )

    def process(self, fuzzresult):
        try:
            dom = xml.dom.minidom.parseString(fuzzresult.history.content)
        except Exception:
            raise FuzzExceptResourceParseError(
                "Error while parsing %s." % fuzzresult.url
            )

        urlList = dom.getElementsByTagName("loc")
        for url in urlList:
            u = url.childNodes[0].data

            self.queue_url(u)



from ..helpers.obj_factory import ObjectFactory

from ..fuzzobjects import FuzzPlugin, FuzzError
from ..factories.fuzzresfactory import resfactory


class PluginFactory(ObjectFactory):
    def __init__(self):
        ObjectFactory.__init__(
            self,
            {
                "plugin_from_recursion": PluginRecursiveBuilder(),
                "plugin_from_error": PluginErrorBuilder(),
                "plugin_from_finding": PluginFindingBuilder(),
                "plugin_from_summary": PluginFindingSummaryBuilder(),
            },
        )


class PluginRecursiveBuilder:
    def __call__(self, name, seed, url):
        plugin = FuzzPlugin()
        plugin.source = name
        plugin._exception = None
        plugin._seed = resfactory.create("fuzzres_from_recursion", seed, url)

        return plugin


class PluginErrorBuilder:
    def __call__(self, name, exception):
        plugin = FuzzPlugin()
        plugin.source = name
        plugin.issue = "Exception within plugin %s: %s" % (name, str(exception))
        plugin._exception = FuzzError(exception)
        plugin._seed = None

        return plugin


class PluginFindingBuilder:
    def __call__(self, name, itype, message, data, severity):
        plugin = FuzzPlugin()
        plugin.source = name
        plugin.issue = message
        plugin.itype = itype
        plugin.data = data
        plugin._exception = None
        plugin._seed = None
        plugin.severity = severity

        return plugin


class PluginFindingSummaryBuilder:
    def __call__(self, message):
        plugin = FuzzPlugin()
        plugin.source = FuzzPlugin.OUTPUT_SOURCE
        plugin.itype = FuzzPlugin.SUMMARY_ITYPE
        plugin.severity = FuzzPlugin.NONE
        plugin._exception = None
        plugin.data = None
        plugin._seed = None
        plugin.issue = message

        return plugin


plugin_factory = PluginFactory()



from .helpers.file_func import get_home, get_path, get_config_dir
from .helpers.obj_factory import Singleton
from . import __version__ as version
from .externals.moduleman.registrant import MulRegistrant
from .externals.moduleman.loader import DirLoader
from .externals.settings.settings import SettingsBase
from .exception import FuzzExceptNoPluginError, FuzzExceptPluginLoadError

import os


ERROR_CODE = -1
BASELINE_CODE = -2


class Settings(SettingsBase):
    def get_config_file(self):
        config_file = "wfuzz.ini"

        config = os.path.join(get_config_dir(check=False), config_file)
        legacy_config = os.path.join(get_home(check=False), config_file)

        if os.path.exists(config):
            return config
        elif os.path.exists(legacy_config):
            return legacy_config
        return os.path.join(get_config_dir(check=True), config_file)

    def set_defaults(self):
        return dict(
            plugins=[("bing_apikey", ""), ("shodan_apikey", "")],
            kbase=[
                (
                    "discovery.blacklist",
                    ".svg-.css-.js-.jpg-.gif-.png-.jpeg-.mov-.avi-.flv-.ico",
                )
            ],
            connection=[
                ("concurrent", "10"),
                ("conn_delay", "90"),
                ("req_delay", "90"),
                ("retries", "3"),
                ("User-Agent", "Wfuzz/%s" % version),
            ],
            general=[
                ("default_printer", "raw"),
                ("cancel_on_plugin_except", "0"),
                ("concurrent_plugins", "3"),
                ("lookup_dirs", "."),
                ("encode_space", "1"),
            ],
        )


class MyRegistrant(MulRegistrant):
    def get_plugin(self, identifier):
        try:
            return MulRegistrant.get_plugin(self, identifier)
        except KeyError as e:
            raise FuzzExceptNoPluginError(
                "Requested plugin %s. Error: %s" % (identifier, str(e))
            )


class Facade(metaclass=Singleton):
    def __init__(self):

        self.__plugins = dict(
            printers=None, scripts=None, encoders=None, iterators=None, payloads=None,
        )

        self.sett = Settings()

    def _load(self, cat):
        try:
            if cat not in self.__plugins:
                raise FuzzExceptNoPluginError("Non-existent plugin category %s" % cat)

            if not self.__plugins[cat]:
                loader_list = []
                loader_list.append(
                    DirLoader(**{"base_dir": cat, "base_path": get_path("../plugins")})
                )
                loader_list.append(
                    DirLoader(**{"base_dir": cat, "base_path": get_home()})
                )
                self.__plugins[cat] = MyRegistrant(loader_list)

            return self.__plugins[cat]
        except Exception as e:
            raise FuzzExceptPluginLoadError("Error loading plugins: %s" % str(e))

    def proxy(self, which):
        return self._load(which)

    def get_registrants(self):
        return self.__plugins.keys()

    def __getattr__(self, name):
        if name in ["printers", "payloads", "iterators", "encoders", "scripts"]:
            return self._load(name)
        else:
            raise AttributeError



import collections

# python 2 and 3
try:
    from itertools import zip_longest
except ImportError:
    from itertools import izip_longest as zip_longest

from queue import PriorityQueue
from threading import Thread, RLock
from .fuzzobjects import FuzzError, FuzzType, FuzzItem

from .exception import FuzzException, FuzzExceptInternalError


class MyPriorityQueue(PriorityQueue):
    def __init__(self, limit=0):
        PriorityQueue.__init__(self, limit)

        self.max_prio = 0

    def _put_priority(self, prio, item, wait):
        self.max_prio = max(prio, self.max_prio)
        PriorityQueue.put(self, (prio, item), wait)

    def put(self, item, wait=True):
        self._put_priority(item.rlevel, item, wait)

    def put_first(self, item, wait=True):
        self._put_priority(0, item, wait)

    def put_last(self, item, wait=True):
        self._put_priority(self.max_prio + 1, item, wait)

    def get(self):
        prio, item = PriorityQueue.get(self, True)

        return item


class FuzzQueue(MyPriorityQueue, Thread):
    def __init__(self, options, queue_out=None, limit=0):
        MyPriorityQueue.__init__(self, limit)
        self.queue_out = queue_out
        self.duplicated = False
        self.syncq = None

        self.stats = options.get("compiled_stats")
        self.options = options

        Thread.__init__(self)
        self.setName(self.get_name())

    def next_queue(self, q):
        self.queue_out = q

    def process(self, item):
        raise NotImplementedError

    def get_name(self):
        raise NotImplementedError

    def process_discarded(self):
        return False

    def items_to_process(self):
        return [FuzzType.RESULT]

    # Override this method if needed. This will be called just before cancelling the job.
    def cancel(self):
        pass

    # Override this method if needed. This will be called just before starting the job.
    def mystart(self):
        pass

    def set_syncq(self, q):
        self.syncq = q

    def qstart(self):
        self.mystart()
        self.start()

    def send_first(self, item):
        self.queue_out.put_first(item)

    def send_last(self, item):
        self.queue_out.put_last(item)

    def qout_join(self):
        self.queue_out.join()

    def send(self, item):
        self.queue_out.put(item)

    def discard(self, item):
        item.discarded = True
        self.send(item)

    def join(self):
        MyPriorityQueue.join(self)

    def tjoin(self):
        Thread.join(self)

    # Override this method if needed. This will be called after job's thread dies.
    def _cleanup(self):
        pass

    def _throw(self, e):
        self.syncq.put_first(FuzzError(e))

    def get_stats(self):
        return {self.get_name(): self.qsize()}

    def run(self):
        cancelling = False

        while 1:
            item = self.get()

            try:
                if item is None:
                    if not self.duplicated:
                        self.send_last(None)
                    self.task_done()
                    break
                elif cancelling:
                    self.task_done()
                    continue
                elif item.item_type == FuzzType.STARTSEED:
                    self.stats.mark_start()
                elif item.item_type == FuzzType.ENDSEED:
                    if not self.duplicated:
                        self.send_last(item)
                    self.task_done()
                    continue
                elif item.item_type == FuzzType.CANCEL:
                    cancelling = True
                    self.send_first(item)
                    self.task_done()
                    continue

                if (
                    not item.discarded or (item.discarded and self.process_discarded())
                ) and item.item_type in self.items_to_process():
                    self.process(item)
                else:
                    self.send(item)

                self.task_done()
            except Exception as e:
                self.task_done()
                self._throw(e)

        self._cleanup()


class LastFuzzQueue(FuzzQueue):
    def __init__(self, options, queue_out=None, limit=0):
        FuzzQueue.__init__(self, options, queue_out, limit)

    def get_name(self):
        return "LastFuzzQueue"

    def process(self):
        pass

    def _cleanup(self):
        pass

    def _throw(self, e):
        self.queue_out.put_first(FuzzError(e))

    def run(self):
        cancelling = False

        while 1:
            item = self.get()

            try:
                self.task_done()

                if item is None:
                    break
                elif cancelling:
                    continue
                elif item.item_type == FuzzType.ERROR:
                    self.qmanager.cancel()
                    self.send_first(item)
                    continue
                elif item.item_type == FuzzType.CANCEL:
                    cancelling = True
                    continue

                if item.item_type == FuzzType.RESULT and not item.discarded:
                    self.send(item)

                if item.item_type == FuzzType.ENDSEED:
                    self.stats.pending_seeds.dec()
                elif item.item_type == FuzzType.RESULT:
                    self.stats.processed.inc()
                    self.stats.pending_fuzz.dec()
                    if item.discarded:
                        self.stats.filtered.inc()

                if self.stats.pending_fuzz() == 0 and self.stats.pending_seeds() == 0:
                    self.qmanager.cleanup()

            except Exception as e:
                self._throw(e)
                self.qmanager.cancel()

        self._cleanup()


class FuzzListQueue(FuzzQueue):
    def __init__(self, options, queues_out, limit=0):
        FuzzQueue.__init__(self, options, queues_out, limit)

        # not to propagate a None/Exception to various queueas at the same level, only propagate through one queue
        for q in self.queue_out[1:]:
            q.duplicated = True

    def set_syncq(self, q):
        for q in self.queue_out:
            q.syncq = q

    def qstart(self):
        for q in self.queue_out:
            q.mystart()
            q.start()
        self.start()

    def send_first(self, item):
        for q in self.queue_out:
            q.put_first(item)

    def send_last(self, item):
        for q in self.queue_out:
            q.put_last(item)

    def send(self, item):
        for q in self.queue_out:
            q.put(item)

    def qout_join(self):
        for q in self.queue_out:
            q.join()

    def join(self):
        self.qout_join()
        MyPriorityQueue.join(self)

    def next_queue(self, nextq):
        for qq in self.queue_out:
            qq.next_queue(nextq)

    def get_stats(self):
        stat_list = []

        for qq in self.queue_out:
            stat_list = stat_list + list(qq.get_stats().items())

        stat_list = stat_list + list(FuzzQueue.get_stats(self).items())

        return dict(stat_list)


class FuzzRRQueue(FuzzListQueue):
    def __init__(self, options, queues_out, limit=0):
        FuzzListQueue.__init__(self, options, queues_out, limit)
        self._next_queue = self._get_next_route()

    def send(self, item):
        next(self._next_queue).put(item)

    def _get_next_route(self):
        i = 0
        while 1:
            yield self.queue_out[i]
            i += 1
            i = i % len(self.queue_out)


class QueueManager:
    def __init__(self, options):
        self._queues = collections.OrderedDict()
        self._lastq = None
        self._syncq = None
        self._mutex = RLock()

        self.options = options

    def add(self, name, q):
        self._queues[name] = q

    def bind(self, lastq):
        with self._mutex:

            queue_list = list(self._queues.values())
            self._lastq = lastq

            self._syncq = LastFuzzQueue(self.options, lastq)
            self._syncq.qmanager = self

            for first, second in zip_longest(queue_list[0:-1:1], queue_list[1::1]):
                first.next_queue(second)
                first.set_syncq(self._syncq)

            queue_list[-1].next_queue(self._syncq)
            queue_list[-1].set_syncq(self._syncq)

    def __getitem__(self, key):
        return self._queues[key]

    def join(self, remove=False):
        with self._mutex:
            for k, q in list(self._queues.items()):
                q.join()
                if remove:
                    del self._queues[k]

    def start(self):
        with self._mutex:
            if self._queues:
                self._syncq.qstart()
                for q in list(self._queues.values()):
                    q.qstart()

                list(self._queues.values())[0].put_first(FuzzItem(FuzzType.STARTSEED))

    def cleanup(self):
        with self._mutex:
            if self._queues:
                list(self._queues.values())[0].put_last(None)
                self.join(remove=True)
                self.options.get("compiled_stats").mark_end()
                self._lastq.put_last(None, wait=False)

                self._queues = collections.OrderedDict()
                self._lastq = None

    def cancel(self):
        with self._mutex:
            if self._queues:
                # stop processing pending items
                for q in list(self._queues.values()):
                    q.cancel()
                    q.put_first(FuzzItem(FuzzType.CANCEL))

                # wait for cancel to be processed
                self.join()

                # send None to stop (almost nicely)
                self.cleanup()

    def get_stats(self):
        stat_list = []

        for q in list(self._queues.values()):
            stat_list = stat_list + list(q.get_stats().items())

        return dict(stat_list)



# Python 2 and 3
try:
    from urllib.parse import urljoin
except ImportError:
    from urlparse import urljoin

from wfuzz.plugin_api.urlutils import check_content_type

from wfuzz.plugin_api.mixins import DiscoveryPluginMixin
from wfuzz.plugin_api.base import BasePlugin
from wfuzz.externals.moduleman.plugin import moduleman_plugin

# Entries format based on:
# http://docstore.mik.ua/orelly/other/cvs/cvs-CHP-6-SECT-9.htm
# Good example at http://webscantest.com/CVS/Entries


@moduleman_plugin
class cvs_extractor(BasePlugin, DiscoveryPluginMixin):
    name = "cvs_extractor"
    author = ("Xavi Mendez (@xmendez)",)
    version = "0.1"
    summary = "Parses CVS/Entries file."
    description = ("Parses CVS/Entries file and enqueues found entries",)
    category = ["active", "discovery"]
    priority = 99
    parameters = ()

    def __init__(self):
        BasePlugin.__init__(self)

    def validate(self, fuzzresult):
        return (
            fuzzresult.url.find("CVS/Entries") >= 0
            and fuzzresult.code == 200
            and check_content_type(fuzzresult, "text")
        )

    def process(self, fuzzresult):
        base_url = urljoin(fuzzresult.url, "..")

        for line in fuzzresult.history.content.splitlines():
            record = line.split("/")
            if len(record) == 6 and record[1]:
                self.queue_url(urljoin(base_url, record[1]))

                # Directory
                if record[0] == "D":
                    self.queue_url(urljoin(base_url, record[1]))
                    self.queue_url(urljoin(base_url, "%s/CVS/Entries" % (record[1])))



from ..fuzzobjects import FPayloadManager, FuzzWord, FuzzWordType

from ..helpers.obj_factory import ObjectFactory, SeedBuilderHelper


class PayManFactory(ObjectFactory):
    def __init__(self):
        ObjectFactory.__init__(
            self,
            {
                "payloadman_from_baseline": BaselinePayloadManBuilder(),
                "payloadman_from_request": FuzzReqPayloadManBuilder(),
                "empty_payloadman": OnePayloadManBuilder(),
            },
        )


class FuzzReqPayloadManBuilder:
    def __call__(self, freq):
        fpm = FPayloadManager()

        for pdict in [
            pdict
            for pdict in SeedBuilderHelper.get_marker_dict(freq)
            if pdict["word"] is not None
        ]:
            fpm.add(pdict)

        return fpm


class OnePayloadManBuilder:
    def __call__(self, content):
        fpm = FPayloadManager()
        fpm.add(
            {"full_marker": None, "word": None, "index": None, "field": None}, content
        )

        return fpm


class BaselinePayloadManBuilder:
    def __call__(self, freq):
        fpm = FPayloadManager()

        for pdict in [
            pdict
            for pdict in SeedBuilderHelper.get_marker_dict(freq)
            if pdict["bl_value"] is not None
        ]:
            fpm.add(pdict, FuzzWord(pdict["bl_value"], FuzzWordType.WORD), True)

        return fpm


payman_factory = PayManFactory()



import pickle as pickle
import gzip

from wfuzz.externals.moduleman.plugin import moduleman_plugin
from wfuzz.exception import FuzzExceptBadFile
from wfuzz.fuzzobjects import FuzzResult, FuzzWordType
from wfuzz.plugin_api.base import BasePayload
from wfuzz.helpers.obj_dyn import rgetattr


@moduleman_plugin
class wfuzzp(BasePayload):
    name = "wfuzzp"
    author = ("Xavi Mendez (@xmendez)",)
    version = "0.2"
    description = (
        "This payload uses pickle.",
        "Warning: The pickle module is not intended to be secure against erroneous or maliciously constructed data.",
        "Never unpickle data received from an untrusted or unauthenticated source.",
        "See: https://blog.nelhage.com/2011/03/exploiting-pickle/",
    )
    summary = "Returns fuzz results' URL from a previous stored wfuzz session."
    category = ["default"]
    priority = 99

    parameters = (
        ("fn", "", True, "Filename of a valid wfuzz result file."),
        (
            "attr",
            None,
            False,
            "Attribute of fuzzresult to return. If not specified the whole object is returned.",
        ),
    )

    default_parameter = "fn"

    def __init__(self, params):
        BasePayload.__init__(self, params)

        self.__max = -1
        self.attr = self.params["attr"]
        self._it = self._gen_wfuzz(self.params["fn"])

    def count(self):
        return self.__max

    def get_next(self):
        next_item = next(self._it)

        return next_item if not self.attr else rgetattr(next_item, self.attr)

    def get_type(self):
        return FuzzWordType.FUZZRES if not self.attr else FuzzWordType.WORD

    def _gen_wfuzz(self, output_fn):
        try:
            with gzip.open(self.find_file(output_fn), "r+b") as output:
                while 1:
                    item = pickle.load(output)
                    if not isinstance(item, FuzzResult):
                        raise FuzzExceptBadFile(
                            "Wrong wfuzz payload format, the object read is not a valid fuzz result."
                        )

                    yield item
        except IOError as e:
            raise FuzzExceptBadFile("Error opening wfuzz payload file. %s" % str(e))
        except EOFError:
            return



import os
import sys
import re
import pkg_resources

from chardet.universaldetector import UniversalDetector
import chardet

from ..exception import FuzzExceptInternalError


def get_filter_help_file():
    FILTER_HELP_FILE = "advanced.rst"
    FILTER_HELP_DEV_FILE = "../../../docs/user/advanced.rst"

    filter_help_text = None
    try:
        fname = pkg_resources.resource_filename("wfuzz", FILTER_HELP_FILE)
        filter_help_text = open(fname).read()
    except IOError:
        filter_help_text = open(get_path(FILTER_HELP_DEV_FILE)).read()

    return filter_help_text


def create_dir(dir_path):
    if not os.path.exists(dir_path):
        os.makedirs(dir_path)


def get_home(check=False, directory=None):
    path = os.path.join(os.path.expanduser("~"), ".wfuzz")
    if check:
        create_dir(path)

    return os.path.join(path, directory) if directory else path


def get_config_dir(check=False):
    config_dir = os.environ.get("XDG_CONFIG_HOME") or os.path.join(
        os.path.expanduser("~"), ".config"
    )
    wfuzz_config_dir = os.path.join(config_dir, "wfuzz")
    if check:
        create_dir(wfuzz_config_dir)
    return wfuzz_config_dir


def get_path(directory=None):
    abspath = os.path.abspath(__file__)
    ret = os.path.dirname(abspath)

    return os.path.join(ret, directory) if directory else ret


def find_file_in_paths(name, path):
    for root, dirs, files in os.walk(path):
        if name in files:
            return os.path.join(root, name)

    return None


class FileDetOpener:
    typical_encodings = [
        "UTF-8",
        "ISO-8859-1",
        "Windows-1251",
        "Shift JIS",
        "Windows-1252",
        "GB2312",
        "EUC-KR",
        "EUC-JP",
        "GBK",
        "ISO-8859-2",
        "Windows-1250",
        "ISO-8859-15",
        "Windows-1256",
        "ISO-8859-9",
        "Big5",
        "Windows-1254",
    ]

    def __init__(self, file_path, encoding=None):
        self.cache = []
        self.file_des = open(file_path, mode="rb")
        self.det_encoding = encoding
        self.encoding_forced = False

    def close(self):
        self.file_des.close()

    def reset(self):
        self.file_des.seek(0)

    def __iter__(self):
        return self

    def __next__(self):
        decoded_line = None
        line = None
        last_error = None

        while decoded_line is None:

            while self.det_encoding is None:
                detect_encoding = self.detect_encoding().get("encoding", "utf-8")
                self.det_encoding = (
                    detect_encoding if detect_encoding is not None else "utf-8"
                )

            if line is None:
                if self.cache:
                    line = self.cache.pop()
                else:
                    line = next(self.file_des)
                    if not line:
                        raise StopIteration

            try:
                decoded_line = line.decode(self.det_encoding)
            except UnicodeDecodeError:
                if last_error is not None and last_error:
                    self.det_encoding = last_error.pop()
                elif last_error is None and not self.encoding_forced:
                    last_error = list(reversed(self.typical_encodings))
                    last_error.append(chardet.detect(line).get("encoding"))
                elif not last_error:
                    raise FuzzExceptInternalError("Unable to decode wordlist file!")

                decoded_line = None

        return decoded_line

    def detect_encoding(self):
        detector = UniversalDetector()
        detector.reset()

        for line in self.file_des:
            detector.feed(line)
            self.cache.append(line)
            if detector.done:
                break

        detector.close()

        return detector.result

    next = __next__  # for Python 2


def open_file_detect_encoding(file_path):
    def detect_encoding(file_path):
        detector = UniversalDetector()
        detector.reset()

        with open(file_path, mode="rb") as file_to_detect:
            for line in file_to_detect:
                detector.feed(line)
                if detector.done:
                    break
        detector.close()

        return detector.result

    if sys.version_info >= (3, 0):
        return open(
            file_path, "r", encoding=detect_encoding(file_path).get("encoding", "utf-8")
        )
    else:
        return open(file_path, "r")



from wfuzz.externals.moduleman.plugin import moduleman_plugin
from wfuzz.plugin_api.payloadtools import ShodanIter
from wfuzz.plugin_api.base import BasePayload
from wfuzz.fuzzobjects import FuzzWordType


@moduleman_plugin
class shodanp(BasePayload):
    name = "shodanp"
    author = ("Xavi Mendez (@xmendez)",)
    version = "0.1"
    description = ("Queries the Shodan API",)

    summary = "Returns URLs of a given Shodan API search (needs api key)."
    category = ["default"]
    priority = 99

    parameters = (
        ("search", "", True, "Shodan search string."),
        ("page", "0", False, "Offset page, starting at zero."),
        (
            "limit",
            "0",
            False,
            "Number of pages (1 query credit = 100 results). Zero for all.",
        ),
    )

    default_parameter = "search"

    def __init__(self, params):
        BasePayload.__init__(self, params)

        search = params["search"]
        page = int(params["page"])
        limit = int(params["limit"])

        self._it = ShodanIter(search, page, limit)

    def count(self):
        return -1

    def close(self):
        self._it._stop()

    def get_type(self):
        return FuzzWordType.WORD

    def get_next(self):
        match = next(self._it)

        port = match["port"]
        scheme = "https" if "ssl" in match or port == 443 else "http"

        if match["hostnames"]:
            for hostname in match["hostnames"]:
                return "{}://{}:{}".format(scheme, hostname, port)
        else:
            return "{}://{}:{}".format(scheme, match["ip_str"], port)



# Python 2 and 3
try:
    from urllib.parse import urljoin
except ImportError:
    from urlparse import urljoin

from wfuzz.plugin_api.mixins import DiscoveryPluginMixin
from wfuzz.plugin_api.base import BasePlugin
from wfuzz.externals.moduleman.plugin import moduleman_plugin


@moduleman_plugin
class svn_extractor(BasePlugin, DiscoveryPluginMixin):
    name = "svn_extractor"
    author = ("Xavi Mendez (@xmendez)",)
    version = "0.1"
    summary = "Parses .svn/entries file."
    description = ("Parses CVS/Entries file and enqueues found entries",)
    category = ["active", "discovery"]
    priority = 99

    parameters = ()

    def __init__(self):
        BasePlugin.__init__(self)

    def validate(self, fuzzresult):
        return fuzzresult.url.find(".svn/entries") > 0 and fuzzresult.code == 200

    def readsvn(self, content):
        """
        Function shamesly copied (and adapted) from https://github.com/anantshri/svn-extractor/
        Credit (C) Anant Shrivastava http://anantshri.info
        """
        old_line = ""
        file_list = []
        dir_list = []
        author_list = []

        for a in content.splitlines():
            # below functionality will find all usernames from svn entries file
            if a == "has-props":
                if old_line not in author_list:
                    author_list.append(old_line)
            if a == "file":
                if old_line not in file_list:
                    file_list.append(old_line)
            if a == "dir":
                if old_line != "":
                    dir_list.append(old_line)
            old_line = a
        return file_list, dir_list, author_list

    def process(self, fuzzresult):
        base_url = fuzzresult.url

        file_list, dir_list, author_list = self.readsvn(fuzzresult.history.content)

        if author_list:
            self.add_result("authors", "SVN authors", ", ".join(author_list))

        for f in file_list:
            u = urljoin(base_url.replace("/.svn/", "/"), f)
            self.queue_url(u)

        for d in dir_list:
            self.queue_url(
                urljoin(base_url.replace("/.svn/", "/"), d) + "/.svn/entries"
            )



# Covered by GPL V2.0
# Coded by Carlos del Ojo Elias (deepbit@gmail.com)
# Lately maintained by Xavi Mendez (xmendez@edge-security.com)

# Python 2 and 3
import sys

if sys.version_info >= (3, 0):
    from urllib.parse import urlparse
    from urllib.parse import urlunparse
else:
    from urlparse import urlparse
    from urlparse import urlunparse

import re
import pycurl

from .Variables import VariablesSet
from .exceptions import ReqRespException
from .Response import Response

from wfuzz.helpers.str_func import python2_3_convert_to_unicode
from wfuzz.helpers.obj_dic import CaseInsensitiveDict

from .TextParser import TextParser


PYCURL_PATH_AS_IS = True
if not hasattr(pycurl, "PATH_AS_IS"):
    PYCURL_PATH_AS_IS = False


class Request:
    def __init__(self):
        self.__host = None  # www.google.com:80
        self.__path = None  # /index.php
        self.__params = None  # Mierdaza de index.php;lskjflkasjflkasjfdlkasdf?
        self.schema = "http"  # http

        # #### Variables calculadas por getters NO SE PUEDEN MODIFICAR
        # self.urlWithoutPath                    # http://www.google.es
        # self.pathWithVariables                        # /index.php?a=b&c=d
        # self.urlWithoutVariables=None                         # http://www.google.es/index.php
        # self.completeUrl=""                                   # http://www.google.es/index.php?a=b
        # self.finalUrl=""                                      # Url despues de hacer el FollowLocation
        # self.redirectUrl=""                                   # Url redirected
        # self.postdata=""              # Datos por POST, toto el string
        # ###############

        self.ContentType = (
            "application/x-www-form-urlencoded"  # None es normal encoding
        )
        self.multiPOSThead = {}

        self.__variablesGET = VariablesSet()
        self._variablesPOST = VariablesSet()
        self._non_parsed_post = None

        # diccionario, por ejemplo headers["Cookie"]
        self._headers = CaseInsensitiveDict(
            {
                "Content-Type": "application/x-www-form-urlencoded",
                "User-Agent": "Mozilla/4.0 (compatible; MSIE 6.0; Windows NT 5.1; SV1; .NET CLR 1.1)",
            }
        )

        self.response = None  # Apunta a la response que produce dicha request

        # ################## lo de debajo no se deberia acceder directamente

        self.time = None  # 23:00:00
        self.ip = None  # 192.168.1.1
        self._method = None
        self.protocol = "HTTP/1.1"  # HTTP/1.1
        self.__performHead = ""
        self.__performBody = ""

        self.__authMethod = None
        self.__userpass = ""

        self.description = ""  # For temporally store imformation

        self.__proxy = None
        self.proxytype = None
        self.__timeout = None
        self.__totaltimeout = None
        self.__finalurl = ""

        self.followLocation = False
        self.__userpass = ""

        self.totaltime = None

    @property
    def method(self):
        if self._method is None:
            return "POST" if self._non_parsed_post is not None else "GET"

        return self._method

    @method.setter
    def method(self, value):
        if value == "None":
            value = None

        self._method = value

    def setFinalUrl(self, fu):
        self.__finalurl = fu

    def __str__(self):
        str = "[ URL: %s" % (self.completeUrl)
        if self.postdata:
            str += ' - {}: "{}"'.format(self.method, self.postdata)
        if "Cookie" in self._headers:
            str += ' - COOKIE: "%s"' % self._headers["Cookie"]
        str += " ]"
        return str

    def getHost(self):
        return self.__host

    def getXML(self, obj):
        r = obj.createElement("request")
        r.setAttribute("method", self.method)
        url = obj.createElement("URL")
        url.appendChild(obj.createTextNode(self.completeUrl))
        r.appendChild(url)
        if self.postdata:
            pd = obj.createElement("PostData")
            pd.appendChild(obj.createTextNode(self.postdata))
            r.appendChild(pd)
        if "Cookie" in self._headers:
            ck = obj.createElement("Cookie")
            ck.appendChild(obj.createTextNode(self._headers["Cookie"]))
            r.appendChild(ck)

        return r

    def __getattr__(self, name):
        if name == "urlWithoutVariables":
            return urlunparse((self.schema, self.__host, self.__path, "", "", ""))
        elif name == "pathWithVariables":
            return urlunparse(
                ("", "", self.__path, "", self.__variablesGET.urlEncoded(), "")
            )
        elif name == "completeUrl":
            return urlunparse(
                (
                    self.schema,
                    self.__host,
                    self.__path,
                    self.__params,
                    self.__variablesGET.urlEncoded(),
                    "",
                )
            )
        elif name == "finalUrl":
            if self.__finalurl:
                return self.__finalurl
            return self.completeUrl
        elif name == "urlWithoutPath":
            return "%s://%s" % (self.schema, self._headers["Host"])
        elif name == "path":
            return self.__path
        elif name == "postdata":
            if self.ContentType == "application/x-www-form-urlencoded":
                return self._variablesPOST.urlEncoded()
            elif self.ContentType == "multipart/form-data":
                return self._variablesPOST.multipartEncoded()
            elif self.ContentType == "application/json":
                return self._variablesPOST.json_encoded()
            else:
                return self._variablesPOST.urlEncoded()
        else:
            raise AttributeError

    def setUrl(self, urltmp):
        self.__variablesGET = VariablesSet()
        self.schema, self.__host, self.__path, self.__params, variables, f = urlparse(
            urltmp
        )
        if "Host" not in self._headers or (not self._headers["Host"]):
            self._headers["Host"] = self.__host

        if variables:
            self.__variablesGET.parseUrlEncoded(variables)

    # ############## PROXY ##################################
    def getProxy(self):
        return self.__proxy

    def setProxy(self, prox, ptype):
        self.__proxy = prox
        self.proxytype = ptype

    # ############## FOLLOW LOCATION ########################
    def setFollowLocation(self, value):
        self.followLocation = value

    # ############# TIMEOUTS ################################
    def setConnTimeout(self, time):
        self.__timeout = time

    def getConnTimeout(self):
        return self.__timeout

    def setTotalTimeout(self, time):
        self.__totaltimeout = time

    def getTotalTimeout(self):
        return self.__totaltimeout

    # ############# Autenticacion ###########################
    def setAuth(self, method, string):
        self.__authMethod = method
        self.__userpass = string

    def getAuth(self):
        return self.__authMethod, self.__userpass

    # ############# TRATAMIENTO VARIABLES GET & POST #########################

    def existsGETVar(self, key):
        return self.__variablesGET.existsVar(key)

    def existPOSTVar(self, key):
        return self._variablesPOST.existsVar(key)

    def setVariablePOST(self, key, value):
        v = self._variablesPOST.getVariable(key)
        v.update(value)

    #       self._headers["Content-Length"] = str(len(self.postdata))

    def setVariableGET(self, key, value):
        v = self.__variablesGET.getVariable(key)
        v.update(value)

    def getGETVars(self):
        return self.__variablesGET.variables

    def getPOSTVars(self):
        return self._variablesPOST.variables

    def setPostData(self, pd, boundary=None):
        self._non_parsed_post = pd
        self._variablesPOST = VariablesSet()

        try:
            if self.ContentType == "multipart/form-data":
                self._variablesPOST.parseMultipart(pd, boundary)
            elif self.ContentType == "application/json":
                self._variablesPOST.parse_json_encoded(pd)
            else:
                self._variablesPOST.parseUrlEncoded(pd)
        except Exception:
            try:
                self._variablesPOST.parseUrlEncoded(pd)
            except Exception:
                print("Warning: POST parameters not parsed")
                pass

    ############################################################################

    def addHeader(self, key, value):
        self._headers[key] = value

    def delHeader(self, key):
        if key in self._headers:
            del self._headers[key]

    def __getitem__(self, key):
        if key in self._headers:
            return self._headers[key]
        else:
            return ""

    def getHeaders(self):
        header_list = []
        for i, j in self._headers.items():
            header_list += ["%s: %s" % (i, j)]
        return header_list

    def head(self):
        conn = pycurl.Curl()
        conn.setopt(pycurl.SSL_VERIFYPEER, False)
        conn.setopt(pycurl.SSL_VERIFYHOST, 0)
        conn.setopt(pycurl.URL, self.completeUrl)

        conn.setopt(pycurl.NOBODY, True)  # para hacer un pedido HEAD

        conn.setopt(pycurl.WRITEFUNCTION, self.header_callback)
        conn.perform()

        rp = Response()
        rp.parseResponse(self.__performHead)
        self.response = rp

    def createPath(self, newpath):
        """Creates new url from a location header || Hecho para el followLocation=true"""
        if "http" in newpath[:4].lower():
            return newpath

        parts = urlparse(self.completeUrl)
        if "/" != newpath[0]:
            newpath = "/".join(parts[2].split("/")[:-1]) + "/" + newpath

        return urlunparse([parts[0], parts[1], newpath, "", "", ""])

    # pycurl - reqresp conversions
    @staticmethod
    def to_pycurl_object(c, req):

        c.setopt(pycurl.MAXREDIRS, 5)

        c.setopt(pycurl.WRITEFUNCTION, req.body_callback)
        c.setopt(pycurl.HEADERFUNCTION, req.header_callback)

        c.setopt(pycurl.NOSIGNAL, 1)
        c.setopt(pycurl.SSL_VERIFYPEER, False)
        c.setopt(pycurl.SSL_VERIFYHOST, 0)

        if PYCURL_PATH_AS_IS:
            c.setopt(pycurl.PATH_AS_IS, 1)

        c.setopt(pycurl.URL, python2_3_convert_to_unicode(req.completeUrl))

        if req.getConnTimeout():
            c.setopt(pycurl.CONNECTTIMEOUT, req.getConnTimeout())

        if req.getTotalTimeout():
            c.setopt(pycurl.TIMEOUT, req.getTotalTimeout())

        authMethod, userpass = req.getAuth()
        if authMethod or userpass:
            if authMethod == "basic":
                c.setopt(pycurl.HTTPAUTH, pycurl.HTTPAUTH_BASIC)
            elif authMethod == "ntlm":
                c.setopt(pycurl.HTTPAUTH, pycurl.HTTPAUTH_NTLM)
            elif authMethod == "digest":
                c.setopt(pycurl.HTTPAUTH, pycurl.HTTPAUTH_DIGEST)
            c.setopt(pycurl.USERPWD, python2_3_convert_to_unicode(userpass))
        else:
            c.unsetopt(pycurl.USERPWD)

        c.setopt(pycurl.HTTPHEADER, python2_3_convert_to_unicode(req.getHeaders()))

        curl_options = {
            "GET": pycurl.HTTPGET,
            "POST": pycurl.POST,
            "PATCH": pycurl.UPLOAD,
            "HEAD": pycurl.NOBODY,
        }

        for o in curl_options.values():
            c.setopt(o, False)

        if req.method in curl_options:
            c.unsetopt(pycurl.CUSTOMREQUEST)
            c.setopt(curl_options[req.method], True)
        else:
            c.setopt(pycurl.CUSTOMREQUEST, req.method)

        if req._non_parsed_post is not None:
            c.setopt(
                pycurl.POSTFIELDS, python2_3_convert_to_unicode(req._non_parsed_post)
            )

        c.setopt(pycurl.FOLLOWLOCATION, 1 if req.followLocation else 0)

        # proxy = req.getProxy()
        # if proxy is not None:
        #     c.setopt(pycurl.PROXY, python2_3_convert_to_unicode(proxy))
        #     if req.proxytype == "SOCKS5":
        #         c.setopt(pycurl.PROXYTYPE, pycurl.PROXYTYPE_SOCKS5)
        #     elif req.proxytype == "SOCKS4":
        #         c.setopt(pycurl.PROXYTYPE, pycurl.PROXYTYPE_SOCKS4)
        #     req.delHeader("Proxy-Connection")
        # else:
        #     c.setopt(pycurl.PROXY, "")

        return c

    def response_from_conn_object(self, conn, header, body):
        # followlocation
        if conn.getinfo(pycurl.EFFECTIVE_URL) != self.completeUrl:
            self.setFinalUrl(conn.getinfo(pycurl.EFFECTIVE_URL))

        self.totaltime = conn.getinfo(pycurl.TOTAL_TIME)

        self.response = Response()
        self.response.parseResponse(header, rawbody=body)

        return self.response

    def perform(self):
        self.__performHead = ""
        self.__performBody = ""
        self.__headersSent = ""

        try:
            conn = Request.to_pycurl_object(pycurl.Curl(), self)
            conn.perform()
            self.response_from_conn_object(conn, self.__performHead, self.__performBody)
        except pycurl.error as error:
            errno, errstr = error
            raise ReqRespException(ReqRespException.FATAL, errstr)
        finally:
            conn.close()

    # ######## ESTE conjunto de funciones no es necesario para el uso habitual de la clase

    def getAll(self):
        pd = self._non_parsed_post if self._non_parsed_post else ""
        string = (
            str(self.method)
            + " "
            + str(self.pathWithVariables)
            + " "
            + str(self.protocol)
            + "\n"
        )
        for i, j in self._headers.items():
            string += i + ": " + j + "\n"
        string += "\n" + pd

        return string

    # #########################################################################

    def header_callback(self, data):
        self.__performHead += data

    def body_callback(self, data):
        self.__performBody += data

    def Substitute(self, src, dst):
        a = self.getAll()
        rx = re.compile(src)
        b = rx.sub(dst, a)
        del rx
        self.parseRequest(b, self.schema)

    def parseRequest(self, rawRequest, prot="http"):
        """ Aun esta en fase BETA y por probar"""
        tp = TextParser()
        tp.setSource("string", rawRequest)

        self._variablesPOST = VariablesSet()
        self._headers = {}  # diccionario, por ejemplo headers["Cookie"]

        tp.readLine()
        try:
            tp.search(r"^(\S+) (.*) (HTTP\S*)$")
            self.method = tp[0][0]
            self.protocol = tp[0][2]
        except Exception as a:
            print(rawRequest)
            raise a

        pathTMP = tp[0][1].replace(" ", "%20")
        pathTMP = ("", "") + urlparse(pathTMP)[2:]
        pathTMP = urlunparse(pathTMP)

        while True:
            tp.readLine()
            if tp.search("^([^:]+): (.*)$"):
                self.addHeader(tp[0][0], tp[0][1])
            else:
                break

        self.setUrl(prot + "://" + self._headers["Host"] + pathTMP)

        # ignore CRLFs until request line
        while tp.lastline == "" and tp.readLine():
            pass

        # TODO: hacky, might need to change tp.readline returning read bytes instead
        pd = ""
        if tp.lastFull_line:
            pd += tp.lastFull_line

        while tp.readLine():
            pd += tp.lastFull_line

        if pd:
            boundary = None
            if "Content-Type" in self._headers:
                values = self._headers["Content-Type"].split(";")
                self.ContentType = values[0].strip().lower()
                if self.ContentType == "multipart/form-data":
                    boundary = values[1].split("=")[1].strip()

            self.setPostData(pd, boundary)



from wfuzz.externals.moduleman.plugin import moduleman_plugin
from wfuzz.plugin_api.base import BasePayload
from wfuzz.fuzzobjects import FuzzWordType


@moduleman_plugin
class names(BasePayload):
    name = "names"
    author = (
        "Christian Martorella",
        "Adapted to newer versions Xavi Mendez (@xmendez)",
    )
    version = "0.1"
    description = ("ie. jon-smith",)
    summary = "Returns possible usernames by mixing the given words, separated by -, using known typical constructions."
    category = ["default"]
    priority = 99

    parameters = (("name", "", True, "Name and surname in the form of name-surname."),)

    default_parameter = "name"

    def __init__(self, params):
        BasePayload.__init__(self, params)

        possibleusernames = []
        name = ""
        llist = self.params["name"].split("-")

        for x in llist:
            if name == "":
                name = name + x
            else:
                name = name + " " + x

            if " " in name:
                parts = name.split()
                possibleusernames.append(parts[0])
                possibleusernames.append(parts[0] + "." + parts[1])
                possibleusernames.append(parts[0] + parts[1])
                possibleusernames.append(parts[0] + "." + parts[1][0])
                possibleusernames.append(parts[0][0] + "." + parts[1])
                possibleusernames.append(parts[0] + parts[1][0])
                possibleusernames.append(parts[0][0] + parts[1])
                str1 = ""
                str2 = ""
                str3 = ""
                str4 = ""
                for i in range(0, len(parts) - 1):
                    str1 = str1 + parts[i] + "."
                    str2 = str2 + parts[i]
                    str3 = str3 + parts[i][0] + "."
                    str4 = str4 + parts[i][0]
                str5 = str1 + parts[-1]
                str6 = str2 + parts[-1]
                str7 = str4 + parts[-1]
                str8 = str3 + parts[-1]
                str9 = str2 + parts[-1][0]
                str10 = str4 + parts[-1][0]
                possibleusernames.append(str5)
                possibleusernames.append(str6)
                possibleusernames.append(str7)
                possibleusernames.append(str8)
                possibleusernames.append(str9)
                possibleusernames.append(str10)
                possibleusernames.append(parts[-1])
                possibleusernames.append(parts[0] + "." + parts[-1])
                possibleusernames.append(parts[0] + parts[-1])
                possibleusernames.append(parts[0] + "." + parts[-1][0])
                possibleusernames.append(parts[0][0] + "." + parts[-1])
                possibleusernames.append(parts[0] + parts[-1][0])
                possibleusernames.append(parts[0][0] + parts[-1])
            else:
                possibleusernames.append(name)

            self.creatednames = possibleusernames
            self.__count = len(possibleusernames)

    def count(self):
        return self.__count

    def get_type(self):
        return FuzzWordType.WORD

    def get_next(self):
        if self.creatednames:
            payl = self.creatednames.pop()
            return payl
        else:
            raise StopIteration



from .exception import FuzzExceptNoPluginError, FuzzExceptBadOptions
from .facade import Facade
from .filters.ppfilter import FuzzResFilterSlice, FuzzResFilter
from .fuzzobjects import FuzzWord, FuzzWordType


class BaseIterator:
    def count(self):
        raise NotImplementedError

    def width(self):
        raise NotImplementedError

    def payloads(self):
        raise NotImplementedError

    def cleanup(self):
        for payload in self.payloads():
            payload.close()


class BaseDictionary:
    def count(self):
        raise NotImplementedError

    def next_word(self):
        raise NotImplementedError

    def __next__(self):
        return self.next_word()

    def __iter__(self):
        return self

    def close(self):
        pass


class EncodeIt(BaseDictionary):
    def __init__(self, parent, encoders_list):
        self.parent = parent
        self.encoders = encoders_list
        self.__generator = self._gen()

    def count(self):
        return self.parent.count() * len(self.encoders)

    def concatenate(self, encoder_name, payload_word):
        string = payload_word.content
        for plugin_name in reversed(encoder_name.split("@")):
            string = Facade().encoders.get_plugin(plugin_name)().encode(string)

        return FuzzWord(string, FuzzWordType.WORD)

    def encode(self, encoder_name, payload_word):
        plugin_list = Facade().encoders.get_plugins(encoder_name)
        if not plugin_list:
            raise FuzzExceptNoPluginError(
                encoder_name
                + " encoder does not exists (-e encodings for a list of available encoders)"
            )

        for plugin_class in plugin_list:
            yield FuzzWord(
                plugin_class().encode(payload_word.content), FuzzWordType.WORD
            )

    def next_word(self):
        return next(self.__generator)

    def _gen(self):
        while 1:
            try:
                payload_word = next(self.parent)
            except StopIteration:
                return

            for encoder_name in self.encoders:
                if encoder_name.find("@") > 0:
                    yield self.concatenate(encoder_name, payload_word)
                else:
                    for string in self.encode(encoder_name, payload_word):
                        yield string

    def __next__(self):
        return next(self.__generator)


class TupleIt(BaseDictionary, BaseIterator):
    def __init__(self, parent):
        self.parent = parent

    def count(self):
        return self.parent.count()

    def width(self):
        return 1

    def payloads(self):
        return [self.parent]

    def next_word(self):
        return (next(self.parent),)


class WrapperIt(BaseDictionary):
    def __init__(self, iterator):
        self._it = iter(iterator)

    def count(self):
        return -1

    def get_type(self):
        return FuzzWordType.WORD

    def next_word(self):
        return FuzzWord(str(next(self._it)), FuzzWordType.WORD)


class SliceIt(BaseDictionary):
    def __init__(self, payload, slicestr):
        self.ffilter = FuzzResFilter(filter_string=slicestr)
        self.ffilter_slice = FuzzResFilterSlice(filter_string=slicestr)
        self.payload = payload

    def count(self):
        return -1

    def get_type(self):
        return self.payload.get_type()

    def _get_filtered_value(self, item):
        if item.type == FuzzWordType.FUZZRES:
            filter_ret = self.ffilter.is_visible(item.content)
        else:
            filter_ret = self.ffilter_slice.is_visible(item.content)

        return filter_ret

    def next_word(self):
        # can be refactored using the walrus operator in python 3.8
        item = next(self.payload)
        filter_ret = self._get_filtered_value(item)

        if not isinstance(filter_ret, bool) and item.type == FuzzWordType.FUZZRES:
            raise FuzzExceptBadOptions(
                "The payload type cannot be modified from FuzzResult to word."
            )

        while isinstance(filter_ret, bool) and not filter_ret:
            item = next(self.payload)
            filter_ret = self._get_filtered_value(item)

        if not isinstance(filter_ret, bool):
            return FuzzWord(filter_ret, item.type)

        return item


class AllVarDictio(BaseDictionary, BaseIterator):
    def __init__(self, iterator, allvar_len):
        self._it = iter(iterator)
        self._count = allvar_len

    def count(self):
        return self._count

    def width(self):
        return 0

    def payloads(self):
        return []

    def next_word(self):
        var_name, fuzz_word = next(self._it)
        return (FuzzWord(var_name, FuzzWordType.WORD), fuzz_word)



from wfuzz.plugin_api.base import BasePlugin
from wfuzz.externals.moduleman.plugin import moduleman_plugin


@moduleman_plugin
class title(BasePlugin):
    name = "title"
    author = ("Xavi Mendez (@xmendez)",)
    version = "0.1"
    summary = "Parses HTML page title"
    description = ("Parses HTML page title",)
    category = ["info", "passive"]
    priority = 99

    parameters = ()

    def __init__(self):
        BasePlugin.__init__(self)

    def validate(self, fuzzresult):
        return True

    def process(self, fuzzresult):
        soup = fuzzresult.history.get_soup()
        title = soup.title.string if soup.title else ""

        if (
            title != ""
            and "title" not in self.kbase
            or title not in self.kbase["title"]
        ):
            self.kbase["title"] = title
            self.add_result("title", "Page title", title)



from wfuzz.externals.moduleman.plugin import moduleman_plugin
from wfuzz.exception import FuzzExceptBadFile
from wfuzz.fuzzobjects import FuzzResult
from wfuzz.fuzzrequest import FuzzRequest
from wfuzz.plugin_api.base import BasePayload
from wfuzz.helpers.obj_dyn import rgetattr
from wfuzz.fuzzobjects import FuzzWordType

import re

import sys

if sys.version_info < (3, 0):
    from io import open

CRLF = "\n"
DELIMITER = "%s%s" % ("=" * 54, CRLF)
CRLF_DELIMITER = CRLF + DELIMITER
HEADER = re.compile(
    r"(\d{1,2}:\d{2}:\d{2} (AM|PM|))[ \t]+(\S+)([ \t]+\[(\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3}|unknown host)\])?"
)


@moduleman_plugin
class burplog(BasePayload):
    name = "burplog"
    author = ("Xavi Mendez (@xmendez)",)
    version = "0.1"
    description = ("Returns fuzz results' URL from a Burp log.",)
    summary = "Returns fuzz results from a Burp log."
    category = ["default"]
    priority = 99

    parameters = (
        ("fn", "", True, "Filename of a valid Burp log file."),
        (
            "attr",
            None,
            False,
            "Attribute of fuzzresult to return. If not specified the whole object is returned.",
        ),
    )

    default_parameter = "fn"

    def __init__(self, params):
        BasePayload.__init__(self, params)

        self.__max = -1
        self.attr = self.params["attr"]
        self._it = self.parse_burp_log(self.params["fn"])

    def count(self):
        return self.__max

    def get_type(self):
        return FuzzWordType.FUZZRES if not self.attr else FuzzWordType.WORD

    def get_next(self):
        next_item = next(self._it)

        return next_item if not self.attr else rgetattr(next_item, self.attr)

    def parse_burp_log(self, burp_log):
        burp_file = None

        try:
            burp_file = open(
                self.find_file(burp_log),
                "r",
                encoding="utf-8",
                errors="surrogateescape",
            )

            history = "START"

            rl = burp_file.readline()
            while rl != "":
                if history == "START":
                    if rl == DELIMITER:
                        history = "HEADER"
                elif history == "HEADER":
                    if rl == DELIMITER:
                        raw_request = ""
                        history = "REQUEST"
                    else:
                        matched = HEADER.match(rl)
                        ctime, host, ip_address = matched.group(1, 3, 5)
                elif history == "REQUEST":
                    if rl == DELIMITER:
                        history = "DELIM1"
                    else:
                        raw_request += rl
                elif history == "DELIM1":
                    if rl == CRLF:
                        raw_response = ""
                        history = "DELIM3"
                    else:
                        raw_response = rl
                        history = "RESPONSE"
                elif history == "RESPONSE":
                    if rl == DELIMITER:
                        history = "DELIM2"
                    else:
                        raw_response += rl
                elif history == "DELIM2":
                    if rl == CRLF:
                        history = "DELIM3"
                elif history == "DELIM3":
                    if rl == CRLF:
                        history = "DELIM4"
                elif history == "DELIM4":
                    if rl == CRLF:
                        fr = FuzzRequest()
                        # last read line contains an extra CRLF
                        fr.update_from_raw_http(
                            raw_request, host[: host.find("://")], raw_response[:-1]
                        )
                        frr = FuzzResult(history=fr)

                        yield frr.update()

                        history = "START"

                rl = burp_file.readline()

        except IOError as e:
            raise FuzzExceptBadFile("Error opening burp log file. %s" % str(e))
        finally:
            if burp_file is not None:
                burp_file.close()



from wfuzz.externals.moduleman.plugin import moduleman_plugin
from wfuzz.exception import FuzzExceptBadFile
from wfuzz.fuzzobjects import FuzzResult, FuzzWordType
from wfuzz.fuzzrequest import FuzzRequest
from wfuzz.plugin_api.base import BasePayload
from wfuzz.helpers.obj_dyn import rgetattr
import xml.etree.cElementTree as ET
from base64 import b64decode


@moduleman_plugin
class burpitem(BasePayload):
    name = "burpitem"
    author = ("Bendegúz Nagy (@PaperTsar)",)
    version = "0.1"
    description = (
        "This payload loads request/response from items saved from Burpsuite.",
    )
    summary = "This payload loads request/response from items saved from Burpsuite."
    category = ["default"]
    priority = 99

    parameters = (
        ("fn", "", True, "Filename of a valid Burp item file."),
        (
            "attr",
            None,
            False,
            "Attribute of fuzzresult to return. If not specified the whole object is returned.",
        ),
    )

    default_parameter = "fn"

    def __init__(self, params):
        BasePayload.__init__(self, params)

        self.__max = -1
        self.attr = self.params["attr"]
        self._it = self._gen_burpitem(self.params["fn"])

    def count(self):
        return self.__max

    def get_next(self):
        next_item = next(self._it)

        return next_item if not self.attr else rgetattr(next_item, self.attr)

    def get_type(self):
        return FuzzWordType.FUZZRES if not self.attr else FuzzWordType.WORD

    def _gen_burpitem(self, output_fn):
        try:
            tree = ET.parse(self.find_file(output_fn))
            for item in tree.getroot().iter("item"):
                fr = FuzzRequest()
                fr.update_from_raw_http(
                    raw=b64decode(item.find("request").text or "").decode("utf-8"),
                    scheme=item.find("protocol").text,
                    raw_response=b64decode(item.find("response").text or ""),
                )
                fr.wf_ip = {
                    "ip": item.find("host").attrib.get("ip", None)
                    or item.find("host").text,
                    "port": item.find("port").text,
                }
                frr = FuzzResult(history=fr)

                yield frr.update()
            return
        except IOError as e:
            raise FuzzExceptBadFile(
                "Error opening Burp items payload file. %s" % str(e)
            )
        except EOFError:
            return



from wfuzz.externals.moduleman.plugin import moduleman_plugin
from wfuzz.dictionaries import BaseIterator

import itertools
from functools import reduce

from builtins import zip as builtinzip


@moduleman_plugin
class zip(BaseIterator):
    name = "zip"
    author = ("Xavi Mendez (@xmendez)",)
    version = "0.1"
    summary = "Returns an iterator that aggregates elements from each of the iterables."
    category = ["default"]
    priority = 99

    def __init__(self, *i):
        self._payload_list = i
        self.__width = len(i)
        self.__count = min([x.count() for x in i])
        self.it = builtinzip(*i)

    def count(self):
        return self.__count

    def width(self):
        return self.__width

    def payloads(self):
        return self._payload_list

    def __next__(self):
        return next(self.it)

    def __iter__(self):
        return self


@moduleman_plugin
class product(BaseIterator):
    name = "product"
    author = ("Xavi Mendez (@xmendez)",)
    version = "0.1"
    summary = "Returns an iterator cartesian product of input iterables."
    category = ["default"]
    priority = 99

    def __init__(self, *i):
        self._payload_list = i
        self.__width = len(i)
        self.__count = reduce(lambda x, y: x * y.count(), i[1:], i[0].count())
        self.it = itertools.product(*i)

    def count(self):
        return self.__count

    def width(self):
        return self.__width

    def payloads(self):
        return self._payload_list

    def __next__(self):
        return next(self.it)

    def __iter__(self):
        return self


@moduleman_plugin
class chain(BaseIterator):
    name = "chain"
    author = ("Xavi Mendez (@xmendez)",)
    version = "0.1"
    summary = "Returns an iterator returns elements from the first iterable until it is exhausted, then proceeds to the next iterable, until all of the iterables are exhausted."
    category = ["default"]
    priority = 99

    def __init__(self, *i):
        self._payload_list = i
        self.__count = sum([x.count() for x in i])
        self.it = itertools.chain(*i)

    def count(self):
        return self.__count

    def width(self):
        return 1

    def payloads(self):
        return self._payload_list

    def __next__(self):
        return (next(self.it),)

    def __iter__(self):
        return self



from ..exception import FuzzExceptIncorrectFilter, FuzzExceptBadOptions
from ..helpers.obj_dyn import (
    rgetattr,
    rsetattr,
)
from ..helpers.str_func import value_in_any_list_item
from ..helpers.obj_dic import DotDict
from ..helpers.utils import diff

import re
import collections
import operator

# Python 2 and 3: alternative 4
try:
    from urllib.parse import unquote
except ImportError:
    from urllib import unquote

from ..facade import Facade, ERROR_CODE


PYPARSING = True
try:
    from pyparsing import (
        Word,
        Group,
        oneOf,
        Optional,
        Suppress,
        ZeroOrMore,
        Literal,
        QuotedString,
        ParseException,
        Regex,
    )
except ImportError:
    PYPARSING = False


class FuzzResFilter:
    FUZZ_MARKER_REGEX = re.compile(r"FUZ\d*Z", re.MULTILINE | re.DOTALL)

    def __init__(self, filter_string=None):
        self.filter_string = filter_string
        self.baseline = None

        quoted_str_value = QuotedString("'", unquoteResults=True, escChar="\\")
        int_values = Word("0123456789").setParseAction(lambda s, l, t: [int(t[0])])
        error_value = Literal("XXX").setParseAction(self.__compute_xxx_value)

        operator_call = Regex(
            r"\|(?P<operator>(m|d|e|un|u|r|l|sw|gre|gregex|unique|startswith|decode|encode|unquote|replace|lower|upper))"
            r"\((?:(?P<param1>('.*?'|\d+))(?:,(?P<param2>('.*?'|\d+)))?)?\)",
            asMatch=True,
        ).setParseAction(lambda s, l, t: [(l, t[0])])

        fuzz_symbol = Regex(
            r"FUZ(?P<index>\d)*Z(?:\[(?P<field>(\w|_|-|\.)+)\])?", asMatch=True
        ).setParseAction(self._compute_fuzz_symbol)
        res_symbol = Regex(
            r"(description|nres|code|chars|lines|words|md5|content|timer|url|l|w|c|(r|history|plugins)(\w|_|-|\.)*|h)"
        ).setParseAction(self._compute_res_symbol)
        bbb_symbol = Regex(
            r"BBB(?:\[(?P<field>(\w|_|-|\.)+)\])?", asMatch=True
        ).setParseAction(self.__compute_bbb_symbol)

        diff_call = Group(
            Suppress(Literal("|"))
            + Literal("diff")
            + Suppress(Literal("("))
            + (fuzz_symbol | res_symbol | bbb_symbol | int_values | quoted_str_value)
            + Suppress(")")
        )

        fuzz_statement = Group(
            (fuzz_symbol | res_symbol | bbb_symbol | int_values | quoted_str_value)
            + Optional(diff_call | operator_call, None)
        ).setParseAction(self.__compute_res_value)

        operator = oneOf("and or")
        not_operator = Optional(oneOf("not"), "notpresent")

        symbol_expr = Group(
            fuzz_statement
            + oneOf("= == != < > >= <= =~ !~ ~ := =+ =-")
            + (error_value | fuzz_statement)
        ).setParseAction(self.__compute_expr)

        definition = symbol_expr ^ fuzz_statement
        definition_not = not_operator + definition
        definition_expr = definition_not + ZeroOrMore(operator + definition_not)

        nested_definition = Group(Suppress("(") + definition_expr + Suppress(")"))
        nested_definition_not = not_operator + nested_definition

        self.finalformula = (nested_definition_not | definition_expr) + ZeroOrMore(
            operator + (nested_definition_not | definition_expr)
        )

        definition_not.setParseAction(self.__compute_not_operator)
        nested_definition_not.setParseAction(self.__compute_not_operator)
        nested_definition.setParseAction(self.__compute_formula)
        self.finalformula.setParseAction(self.__myreduce)

        self.res = None
        self.stack = []
        self._cache = collections.defaultdict(set)

    def set_baseline(self, res):
        self.baseline = res

    def _compute_res_symbol(self, tokens):
        return self._get_field_value(self.res, tokens[0])

    def _compute_fuzz_symbol(self, tokens):
        match_dict = tokens[0].groupdict()
        p_index = int(match_dict["index"]) if match_dict["index"] is not None else 1
        fuzz_val = None

        try:
            fuzz_val = self.res.payload_man.get_payload_content(p_index)
        except IndexError:
            raise FuzzExceptIncorrectFilter(
                "Non existent FUZZ payload! Use a correct index."
            )

        if match_dict["field"]:
            fuzz_val = self._get_field_value(fuzz_val, match_dict["field"])

        return fuzz_val

    def __compute_res_value(self, tokens):
        fuzz_val, token_tuple = tokens[0]

        if token_tuple:
            location, operator_match = token_tuple

            if location == "diff":
                return diff(operator_match, fuzz_val)
            else:
                if operator_match and operator_match.groupdict()["operator"]:
                    fuzz_val = self._get_operator_value(
                        location, fuzz_val, operator_match.groupdict()
                    )

        if isinstance(fuzz_val, list):
            return [fuzz_val]
        return fuzz_val

    def _get_payload_value(self, p_index):
        try:
            return self.res.payload_man.get_payload_content(p_index)
        except IndexError:
            raise FuzzExceptIncorrectFilter(
                "Non existent FUZZ payload! Use a correct index."
            )

    def _get_field_value(self, fuzz_val, field):
        self.stack.append(field)

        try:
            ret = rgetattr(fuzz_val, field)
        except IndexError:
            raise FuzzExceptIncorrectFilter(
                "Non existent FUZZ payload! Use a correct index."
            )
        except AttributeError as e:
            raise FuzzExceptIncorrectFilter(
                "Attribute {} not found in fuzzresult or using a string payload. {}".format(
                    field, str(e)
                )
            )

        if isinstance(ret, list):
            return [ret]
        return ret

    def __compute_bbb_symbol(self, tokens):
        if self.baseline is None:
            raise FuzzExceptBadOptions(
                "FilterQ: specify a baseline value when using BBB"
            )

        match_dict = tokens[0].groupdict()

        ret = None

        if match_dict["field"]:
            ret = self._get_field_value(self.baseline, match_dict["field"])
        else:
            element = self.stack.pop() if self.stack else None

            if element == "l" or element == "lines":
                ret = self.baseline.lines
            elif element == "c" or element == "code":
                ret = self.baseline.code
            elif element == "w" or element == "words":
                ret = self.baseline.words
            elif element == "h" or element == "chars":
                return self.baseline.chars
            elif element == "index" or element == "i":
                ret = self.baseline.nres
            else:
                ret = self.baseline.payload_man.get_payload_content(1)

        return ret

    def _get_operator_value(self, location, fuzz_val, match_dict):
        op = match_dict["operator"]
        param1 = match_dict["param1"]
        param2 = match_dict["param2"]

        if param1:
            param1 = param1.strip("'")
        if param2:
            param2 = param2.strip("'")

        if (op == "un" or op == "unquote") and param1 is None and param2 is None:
            ret = unquote(fuzz_val)
        elif (op == "e" or op == "encode") and param1 is not None and param2 is None:
            ret = Facade().encoders.get_plugin(param1)().encode(fuzz_val)
        elif (op == "d" or op == "decode") and param1 is not None and param2 is None:
            ret = Facade().encoders.get_plugin(param1)().decode(fuzz_val)
        elif op == "r" or op == "replace":
            return fuzz_val.replace(param1, param2)
        elif op == "upper":
            return fuzz_val.upper()
        elif op == "lower" or op == "l":
            return fuzz_val.lower()
        elif op == "gregex" or op == "gre":
            search_res = None
            try:
                regex = re.compile(param1)
                search_res = regex.search(fuzz_val)
            except re.error as e:
                raise FuzzExceptBadOptions(
                    "Invalid regex expression used in expression: %s" % str(e)
                )

            if search_res is None:
                return ""
            return search_res.group(1)
        elif op == "startswith" or op == "sw":
            return fuzz_val.strip().startswith(param1)
        elif op == "unique" or op == "u":
            if fuzz_val not in self._cache[location]:
                self._cache[location].add(fuzz_val)
                return True
            else:
                return False
        else:
            raise FuzzExceptBadOptions(
                "Bad format, expression should be m,d,e,r,s(value,value)"
            )

        return ret

    def __compute_xxx_value(self, tokens):
        return ERROR_CODE

    def __compute_expr(self, tokens):
        leftvalue, exp_operator, rightvalue = tokens[0]

        # a bit hacky but we don't care about fields in the right hand side of the expression
        if len(self.stack) > 1:
            self.stack.pop()

        field_to_set = self.stack.pop() if self.stack else None

        try:
            if exp_operator in ["=", "=="]:
                return str(leftvalue) == str(rightvalue)
            elif exp_operator == "<=":
                return int(leftvalue) <= int(rightvalue)
            elif exp_operator == ">=":
                return int(leftvalue) >= int(rightvalue)
            elif exp_operator == "<":
                return int(leftvalue) < int(rightvalue)
            elif exp_operator == ">":
                return int(leftvalue) > int(rightvalue)
            elif exp_operator == "!=":
                return leftvalue != rightvalue
            elif exp_operator == "=~":
                regex = re.compile(rightvalue, re.MULTILINE | re.DOTALL)
                return regex.search(leftvalue) is not None
            elif exp_operator in ["!~", "~"]:
                ret = True

                if isinstance(leftvalue, str):
                    ret = rightvalue.lower() in leftvalue.lower()
                elif isinstance(leftvalue, list):
                    ret = value_in_any_list_item(rightvalue, leftvalue)
                elif isinstance(leftvalue, dict) or isinstance(leftvalue, DotDict):
                    ret = rightvalue.lower() in str(leftvalue).lower()
                else:
                    raise FuzzExceptBadOptions(
                        "Invalid operand type {}".format(rightvalue)
                    )

                return ret if exp_operator == "~" else not ret
            elif exp_operator == ":=":
                rsetattr(self.res, field_to_set, rightvalue, None)
            elif exp_operator == "=+":
                rsetattr(self.res, field_to_set, rightvalue, operator.add)
            elif exp_operator == "=-":
                if isinstance(rightvalue, str):
                    rsetattr(self.res, field_to_set, rightvalue, lambda x, y: y + x)
                else:
                    rsetattr(self.res, field_to_set, rightvalue, operator.sub)
        except re.error as e:
            raise FuzzExceptBadOptions(
                "Invalid regex expression used in expression: %s" % str(e)
            )
        except TypeError as e:
            raise FuzzExceptBadOptions(
                "Invalid operand types used in expression: %s" % str(e)
            )
        except ParseException as e:
            raise FuzzExceptBadOptions("Invalid filter: %s" % str(e))

        return True

    def __myreduce(self, elements):
        first = elements[0]
        for i in range(1, len(elements), 2):
            if elements[i] == "and":
                first = first and elements[i + 1]
            elif elements[i] == "or":
                first = first or elements[i + 1]

        self.stack = []

        if isinstance(first, list):
            return [first]
        return first

    def __compute_not_operator(self, tokens):
        operator, value = tokens

        if operator == "not":
            return not value

        if isinstance(value, list):
            return [value]
        return value

    def __compute_formula(self, tokens):
        return self.__myreduce(tokens[0])

    def is_active(self):
        return self.filter_string

    def is_visible(self, res, filter_string=None):
        if filter_string is None:
            filter_string = self.filter_string
        self.res = res
        try:
            return self.finalformula.parseString(filter_string, parseAll=True)[0]
        except ParseException as e:
            raise FuzzExceptIncorrectFilter(
                "Incorrect filter expression, check documentation. {}".format(str(e))
            )
        except AttributeError as e:
            raise FuzzExceptIncorrectFilter(
                "It is only possible to use advanced filters when using a non-string payload. %s"
                % str(e)
            )

    def get_fuzz_words(self):
        fuzz_words = self.FUZZ_MARKER_REGEX.findall(self.filter_string)

        return fuzz_words


class FuzzResFilterSlice(FuzzResFilter):
    # When using slice we don't have previous payload context but directly a word from the dictionary
    def _compute_fuzz_symbol(self, tokens):
        match_dict = tokens[0].groupdict()

        p_index = match_dict["index"] if match_dict["index"] is not None else 1

        if p_index != 1:
            raise FuzzExceptIncorrectFilter(
                "Non existent FUZZ payload! Use a correct index."
            )

        fuzz_val = self.res

        if match_dict["field"]:
            fuzz_val = self._get_field_value(self.res, match_dict["field"])

        return fuzz_val



import re

from wfuzz.plugin_api.base import BasePlugin
from wfuzz.externals.moduleman.plugin import moduleman_plugin


@moduleman_plugin
class listing(BasePlugin):
    name = "listing"
    author = ("Xavi Mendez (@xmendez)",)
    version = "0.1"
    summary = "Looks for directory listing vulnerabilities"
    description = ("Looks for directory listing vulnerabilities",)
    category = ["default", "passive"]
    priority = 99

    parameters = ()

    def __init__(self):
        BasePlugin.__init__(self)

        dir_indexing_regexes = []

        dir_indexing_regexes.append("<title>Index of /")
        dir_indexing_regexes.append('<a href="\\?C=N;O=D">Name</a>')
        dir_indexing_regexes.append("Last modified</a>")
        dir_indexing_regexes.append("Parent Directory</a>")
        dir_indexing_regexes.append("Directory Listing for")
        dir_indexing_regexes.append("<TITLE>Folder Listing.")
        dir_indexing_regexes.append("<TITLE>Folder Listing.")
        dir_indexing_regexes.append('<table summary="Directory Listing" ')
        dir_indexing_regexes.append("- Browsing directory ")
        dir_indexing_regexes.append(
            '">\\[To Parent Directory\\]</a><br><br>'
        )  # IIS 6.0 and 7.0
        dir_indexing_regexes.append(
            '<A HREF=".*?">.*?</A><br></pre><hr></body></html>'
        )  # IIS 5.0

        self.regex = []
        for i in dir_indexing_regexes:
            self.regex.append(re.compile(i, re.MULTILINE | re.DOTALL))

    def validate(self, fuzzresult):
        return fuzzresult.code in [200]

    def process(self, fuzzresult):
        for r in self.regex:
            if len(r.findall(fuzzresult.history.content)) > 0:
                self.add_result("msg", "Directory listing identified", None)
                break



# Python 2 and 3 (after ``pip install configparser``):
try:
    from configparser import ConfigParser
except ImportError:
    import ConfigParser
import os
import sys


class SettingsBase:
    """
    Contains application settings. uses a ConfigParser
    """

    def __init__(self, save=False):
        self.cparser = ConfigParser()

        self.set_all(self.set_defaults())
        self.filename = os.path.join(
            self._path_to_program_dir(), self.get_config_file()
        )
        self.cparser.read(self.filename)

    # Base members should implement

    def get_config_file(self):
        """Returns the name of the file where the config is saved."""
        raise NotImplementedError

    def set_defaults(self):
        """
        Returns a dictionary with the default settings in the form of
        { \
                Section: [ \
                    ("setting_x", '5'),
                    ...
                    ("setting_y", '5'),
                ],
        ...
        }
        """
        raise NotImplementedError

    def has_option(self, section, setting):
        return self.cparser.has_option(section, setting)

    def set(self, section, setting, value):
        self.cparser.set(section, setting, value)

    def get(self, section, setting):
        value = self.cparser.get(section, setting)
        return value

    def get_section(self, section):
        return self.cparser.items(section)

    def get_options(self, section):
        return self.cparser.options(section)

    def get_sections(self):
        return self.cparser.sections()

    def get_all(self):
        sett = {}

        # dump entire config file
        for section in self.cparser.sections():
            for option in self.cparser.options(section):
                if section not in sett:
                    sett[section] = []
                sett[section].append((option, self.cparser.get(section, option)))

        return sett

    def set_all(self, sett):
        self.cparser = ConfigParser()
        for section, settings in sett.items():
            self.cparser.add_section(section)
            for key, value in settings:
                self.cparser.set(section, key, value)

    def save(self):
        try:
            with open(self.filename, "w") as iniFile:
                self.cparser.write(iniFile)
        except Exception:
            return False
        return True

    def _path_to_program_dir(self):
        """
        Returns path to program directory
        """
        path = sys.argv[0]

        if not os.path.isdir(path):
            path = os.path.dirname(path)

        if not path:
            return "."

        return path



from wfuzz.externals.moduleman.plugin import moduleman_plugin
from wfuzz.exception import FuzzExceptPluginBadParams, FuzzExceptBadInstall
from wfuzz.plugin_api.base import BasePayload
from wfuzz.fuzzobjects import FuzzWordType


@moduleman_plugin
class ipnet(BasePayload):
    name = "ipnet"
    author = ("Xavi Mendez (@xmendez)",)
    version = "0.1"
    description = ("ie. 192.168.1.0/24", "Requires: netaddr module")
    summary = "Returns list of IP addresses of a network."
    category = ["default"]
    priority = 99

    parameters = (("net", "", True, "Network range in the form ip/mask."),)

    default_parameter = "net"

    def __init__(self, params):
        BasePayload.__init__(self, params)

        try:
            from netaddr import IPNetwork
            from netaddr.core import AddrFormatError

            net = IPNetwork("%s" % self.params["net"])
            self.f = net.iter_hosts()
            self.__count = net.size - 2

            if self.__count <= 0:
                raise FuzzExceptPluginBadParams(
                    "There are not hosts in the specified network"
                )

        except ValueError:
            raise FuzzExceptPluginBadParams(
                "The specified network has an incorrect format."
            )
        except ImportError:
            raise FuzzExceptBadInstall(
                "ipnet plugin requires netaddr module. Please install it using pip."
            )
        except AddrFormatError:
            raise FuzzExceptPluginBadParams(
                "The specified network has an incorrect format."
            )

    def get_type(self):
        return FuzzWordType.WORD

    def get_next(self):
        return str(next(self.f))

    def count(self):
        return self.__count



from wfuzz.externals.moduleman.plugin import moduleman_plugin

# Python 2 and 3
try:
    from urllib.parse import quote
    from urllib.parse import unquote
except ImportError:
    from urllib import quote
    from urllib import unquote

# Python 2 and 3
try:
    from base64 import decodebytes as b64decode
    from base64 import standard_b64encode
except ImportError:
    from base64 import decodestring as b64decode
    from base64 import standard_b64encode

import re
import binascii
import random
import hashlib
import html


@moduleman_plugin("encode")
class none:
    name = "none"
    author = ("Xavi Mendez (@xmendez)",)
    version = "0.1"
    summary = "Returns string without changes"
    category = ["default"]
    priority = 99

    def encode(self, string):
        return string

    def decode(self, string):
        return string


@moduleman_plugin("encode")
class urlencode:
    name = "urlencode"
    author = (
        "Carlos del Ojo",
        "Christian Martorella",
        "Adapted to newer versions Xavi Mendez (@xmendez)",
    )
    version = "0.1"
    summary = "Replace special characters in string using the %xx escape. Letters, digits, and the characters '_.-' are never quoted."
    category = ["url_safe", "url"]
    priority = 99

    def encode(self, string):
        return quote(string)

    def decode(self, string):
        return unquote(string)


@moduleman_plugin("encode")
class double_urlencode:
    name = "double_urlencode"
    author = (
        "Carlos del Ojo",
        "Christian Martorella",
        "Adapted to newer versions Xavi Mendez (@xmendez)",
    )
    version = "0.1"
    summary = "Applies a double encode to special characters in string using the %25xx escape. Letters, digits, and the characters '_.-' are never quoted."
    category = ["url_safe", "url"]
    priority = 99

    def encode(self, string):
        return quote(quote(string))

    def decode(self, string):
        return unquote(unquote(string))


@moduleman_plugin("encode")
class base64:
    name = "base64"
    author = (
        "Carlos del Ojo",
        "Christian Martorella",
        "Adapted to newer versions Xavi Mendez (@xmendez)",
    )
    version = "0.1"
    summary = "Encodes the given string using base64"
    category = ["hashes"]
    priority = 99

    def encode(self, string):
        return standard_b64encode(string.encode("utf-8")).decode("utf-8")

    def decode(self, string):
        return b64decode(string.encode("utf-8")).decode("utf-8")


@moduleman_plugin("encode")
class uri_triple_hex:
    name = "uri_triple_hex"
    author = (
        "Carlos del Ojo",
        "Christian Martorella",
        "Adapted to newer versions Xavi Mendez (@xmendez)",
    )
    version = "0.1"
    summary = "Encodes ALL charachers using the %25%xx%xx escape."
    category = ["url"]
    priority = 99

    def encode(self, string):
        strt = ""
        s = re.compile(r"/|;|=|:|&|@|\\|\?")
        for c in string:
            if s.search(c):
                strt += c
                continue
            temp = hex(ord(c))[2:]
            strt += "%%25%%%02x%%%02x" % (ord(temp[:1]), ord(temp[1:]))
        return strt


@moduleman_plugin("encode")
class uri_double_hex:
    name = "uri_double_hex"
    author = (
        "Carlos del Ojo",
        "Christian Martorella",
        "Adapted to newer versions Xavi Mendez (@xmendez)",
    )
    version = "0.1"
    summary = "Encodes ALL charachers using the %25xx escape."
    category = ["url"]
    priority = 99

    def encode(self, string):
        strt = ""
        con = "%%25%02x"
        s = re.compile(r"/|;|=|:|&|@|\\|\?")
        for c in string:
            if s.search(c):
                strt += c
                continue
            strt += con % ord(c)
        return strt


@moduleman_plugin("encode")
class uri_hex:
    name = "uri_hex"
    author = (
        "Carlos del Ojo",
        "Christian Martorella",
        "Adapted to newer versions Xavi Mendez (@xmendez)",
    )
    version = "0.1"
    summary = "Encodes ALL charachers using the %xx escape."
    category = ["url"]
    priority = 99

    def encode(self, string):
        strt = ""
        con = "%%%02x"
        s = re.compile(r"/|;|=|:|&|@|\\|\?")
        for c in string:
            if s.search(c):
                strt += c
                continue
            strt += con % ord(c)
        return strt


@moduleman_plugin("encode")
class random_upper:
    name = "random_upper"
    author = (
        "Carlos del Ojo",
        "Christian Martorella",
        "Adapted to newer versions Xavi Mendez (@xmendez)",
    )
    version = "0.1"
    summary = "Replaces random characters in string with its capitals letters"
    category = ["default"]
    priority = 99

    def encode(self, string):
        strt = ""
        for c in string:
            x = int(random.uniform(0, 10))
            x = x % 2
            if x == 1:
                strt += c.upper()
            else:
                strt += c
        return strt


@moduleman_plugin("encode")
class second_nibble_hex:
    name = "second_nibble_hex"
    author = (
        "Carlos del Ojo",
        "Christian Martorella",
        "Adapted to newer versions Xavi Mendez (@xmendez)",
    )
    version = "0.1"
    summary = "Replaces ALL characters in string using the %?%dd escape"
    category = ["url"]
    priority = 99

    def encode(self, string):
        strt = ""
        s = re.compile(r"/|;|=|:|&|@|\\|\?")
        for c in string:
            if s.search(c):
                strt += c
                continue
            temp = hex(ord(c))[2:]
            strt += "%%%s%%%02x" % (str(temp[:1]), ord(temp[1:]))
        return strt


@moduleman_plugin("encode")
class first_nibble_hex:
    name = "first_nibble_hex"
    author = (
        "Carlos del Ojo",
        "Christian Martorella",
        "Adapted to newer versions Xavi Mendez (@xmendez)",
    )
    version = "0.1"
    summary = "Replaces ALL characters in string using the %%dd? escape"
    category = ["url"]
    priority = 99

    def encode(self, string):
        strt = ""
        s = re.compile(r"/|;|=|:|&|@|\\|\?")
        for c in string:
            if s.search(c):
                strt += c
                continue
            temp = hex(ord(c))[2:]
            strt += "%%%%%02x%s" % (ord(temp[:1]), str(temp[1:]))
        return strt


@moduleman_plugin("encode")
class doble_nibble_hex:
    name = "doble_nibble_hex"
    author = (
        "Carlos del Ojo",
        "Christian Martorella",
        "Adapted to newer versions Xavi Mendez (@xmendez)",
    )
    version = "0.1"
    summary = "Replaces ALL characters in string using the %%dd%dd escape"
    category = ["url"]
    priority = 99

    def encode(self, string):
        strt = ""
        fin = ""
        con = "%%%02x"
        s = re.compile(r"/|;|=|:|&|@|\\|\?")
        enc = uri_hex()
        strt = enc.encode(string)
        for c in strt:
            if not c == "%":
                if s.search(c):
                    fin += c
                    continue
                fin += con % ord(c)
            else:
                fin += c
        return fin


@moduleman_plugin("encode")
class sha1:
    name = "sha1"
    summary = "Applies a sha1 hash to the given string"
    author = (
        "Carlos del Ojo",
        "Christian Martorella",
        "Adapted to newer versions Xavi Mendez (@xmendez)",
    )
    version = "0.1"
    category = ["hashes"]
    priority = 99

    def encode(self, string):
        s = hashlib.sha1()
        s.update(string.encode("utf-8"))
        res = s.hexdigest()
        return res


@moduleman_plugin("encode")
class sha256:
    name = "sha256"
    summary = "Applies a sha256 hash to the given string"
    author = ("Dustin Evans (@dustinaevans)",)
    version = "0.1"
    category = ["hashes"]
    priority = 99

    def encode(self, string):
        s = hashlib.sha256()
        s.update(string.encode("utf-8"))
        res = s.hexdigest()
        return res


@moduleman_plugin("encode")
class sha512:
    name = "sha512"
    summary = "Applies a sha512 hash to the given string"
    author = ("Dustin Evans (@dustinaevans)",)
    version = "0.1"
    category = ["hashes"]
    priority = 99

    def encode(self, string):
        s = hashlib.sha512()
        s.update(string.encode("utf-8"))
        res = s.hexdigest()
        return res


@moduleman_plugin("encode")
class md5:
    name = "md5"
    author = (
        "Carlos del Ojo",
        "Christian Martorella",
        "Adapted to newer versions Xavi Mendez (@xmendez)",
    )
    version = "0.1"
    summary = "Applies a md5 hash to the given string"
    category = ["hashes"]
    priority = 99

    def encode(self, string):
        m = hashlib.new("md5")
        m.update(string.encode("utf-8"))
        res = m.hexdigest()
        return res


@moduleman_plugin("encode")
class hexlify:
    name = "hexlify"
    author = (
        "Carlos del Ojo",
        "Christian Martorella",
        "Adapted to newer versions Xavi Mendez (@xmendez)",
    )
    version = "0.1"
    summary = "Every byte of data is converted into the corresponding 2-digit hex representation."
    category = ["default"]
    priority = 99

    def encode(self, string):
        return binascii.hexlify(string.encode("utf-8")).decode("utf-8")

    def decode(self, string):
        return binascii.unhexlify(string.encode("utf-8")).decode("utf-8")


@moduleman_plugin("encode")
class html_escape:
    name = "html_escape"
    author = (
        "Carlos del Ojo",
        "Christian Martorella",
        "Adapted to newer versions Xavi Mendez (@xmendez)",
    )
    version = "0.1"
    summary = 'Convert the characters &<>" in string to HTML-safe sequences.'
    category = ["html"]
    priority = 99

    def encode(self, string):
        return html.escape(string, quote=True)


@moduleman_plugin("encode")
class html_decimal:
    name = "html_decimal"
    author = (
        "Carlos del Ojo",
        "Christian Martorella",
        "Adapted to newer versions Xavi Mendez (@xmendez)",
    )
    version = "0.1"
    summary = "Replaces ALL characters in string using the &#dd; escape"
    category = ["html"]
    priority = 99

    def encode(self, string):
        new = ""
        for x in string:
            new += "&#" + str(ord(x)) + ";"
        return new


@moduleman_plugin("encode")
class html_hexadecimal:
    name = "html_hexadecimal"
    author = (
        "Carlos del Ojo",
        "Christian Martorella",
        "Adapted to newer versions Xavi Mendez (@xmendez)",
    )
    version = "0.1"
    summary = "Replaces ALL characters in string using the &#xx; escape"
    category = ["html"]
    priority = 99

    def encode(self, string):
        new = ""
        for x in string:
            val = "%02x" % ord(x)
            new += "&#x" + str(val) + ";"
        return new


@moduleman_plugin("encode")
class utf8_binary:
    name = "utf8_binary"
    author = (
        "Carlos del Ojo",
        "Christian Martorella",
        "Adapted to newer versions Xavi Mendez (@xmendez)",
    )
    version = "0.1"
    summary = "Replaces ALL characters in string using the \\uxx escape"
    category = ["url"]
    priority = 99

    def encode(self, string):
        new = ""
        for x in string:
            val = "%02x" % ord(x)
            new += "\\x" + str(val)
        return new


@moduleman_plugin("encode")
class utf8:
    name = "utf8"
    author = (
        "Carlos del Ojo",
        "Christian Martorella",
        "Adapted to newer versions Xavi Mendez (@xmendez)",
    )
    version = "0.1"
    summary = "Replaces ALL characters in string using the \\u00xx escape"
    category = ["url"]
    priority = 99

    def encode(self, string):
        new = ""
        for x in string:
            val = "%02x" % ord(x)
            if len(val) == 2:
                new += "\\u00" + str(val)
            else:
                new += "\\u" + str(val)
        return new


@moduleman_plugin("encode")
class uri_unicode:
    name = "uri_unicode"
    author = (
        "Carlos del Ojo",
        "Christian Martorella",
        "Adapted to newer versions Xavi Mendez (@xmendez)",
    )
    version = "0.1"
    summary = "Replaces ALL characters in string using the %u00xx escape"
    category = ["url"]
    priority = 99

    def encode(self, string):
        new = ""
        for x in string:
            val = "%02x" % ord(x)
            if len(val) == 2:
                new += "%u00" + str(val)
            else:
                new += "%u" + str(val)
        return new


@moduleman_plugin("encode")
class mysql_char:
    name = "mysql_char"
    author = (
        "Carlos del Ojo",
        "Christian Martorella",
        "Adapted to newer versions Xavi Mendez (@xmendez)",
    )
    version = "0.1"
    summary = "Converts ALL characters to MySQL's char(xx)"
    category = ["db"]
    priority = 99

    def encode(self, string):
        new = "CHAR("
        for x in string:
            val = str(ord(x))
            new += str(val) + ","
        new = new.strip(",")
        new += ")"
        return new

    def decode(self, string):
        temp = string.strip("CHAR").strip("(").strip(")").split(",")
        new = ""
        for x in temp:
            new += chr(int(x))
        return new


@moduleman_plugin("encode")
class mssql_char:
    name = "mssql_char"
    author = (
        "Carlos del Ojo",
        "Christian Martorella",
        "Adapted to newer versions Xavi Mendez (@xmendez)",
    )
    version = "0.1"
    summary = "Converts ALL characters to MsSQL's char(xx)"
    category = ["db"]
    priority = 99

    def encode(self, string):
        new = ""
        for x in string:
            val = str(ord(x))
            new += "CHAR(" + str(val) + ")+"
        new = new.strip("+")
        return new

    def decode(self, string):
        new = ""
        temp = string.split("+")
        for x in temp:
            x = x.strip("CHAR").strip(")").strip("(")
            new += chr(int(x))
        return new


@moduleman_plugin("encode")
class oracle_char:
    name = "oracle_char"
    author = (
        "Carlos del Ojo",
        "Christian Martorella",
        "Adapted to newer versions Xavi Mendez (@xmendez)",
    )
    version = "0.1"
    summary = "Converts ALL characters to Oracle's chr(xx)"
    category = ["db"]
    priority = 99

    def encode(self, string):
        new = ""
        for x in string:
            val = str(ord(x))
            new += "chr(" + val + ")||"
        new = new.strip("||")
        return new

    def decode(self, string):
        new = ""
        temp = string.split("||")
        for x in temp:
            x = x.strip("chr").strip(")").strip("(")
            new += chr(int(x))
        return new



from wfuzz.externals.moduleman.plugin import moduleman_plugin
from wfuzz.plugin_api.base import BasePayload
from wfuzz.fuzzobjects import FuzzWordType

import sys


@moduleman_plugin
class stdin(BasePayload):
    name = "stdin"
    author = ("Xavi Mendez (@xmendez)",)
    version = "0.1"
    description = ()
    summary = "Returns each item read from stdin."
    category = ["default"]
    priority = 99

    parameters = ()

    default_parameter = ""

    def __init__(self, params):
        BasePayload.__init__(self, params)
        self.__count = -1

    def count(self):
        return self.__count

    def get_type(self):
        return FuzzWordType.WORD

    def get_next(self):
        line = next(sys.stdin).strip()

        return line



import pycurl

# Python 2 and 3
import sys

if sys.version_info >= (3, 0):
    from urllib.parse import urlparse
else:
    from urlparse import urlparse

from collections import namedtuple

from .externals.reqresp import Request, Response
from .exception import FuzzExceptBadAPI, FuzzExceptBadOptions
from .facade import Facade
from .mixins import FuzzRequestUrlMixing, FuzzRequestSoupMixing

from .helpers.str_func import python2_3_convert_from_unicode
from .helpers.obj_dic import DotDict


class headers(object):
    class header(DotDict):
        def __str__(self):
            return "\n".join(["{}: {}".format(k, v) for k, v in self.items()])

    def __init__(self, req):
        self._req = req

    @property
    def response(self):
        return (
            headers.header(self._req.response.getHeaders())
            if self._req.response
            else headers.header()
        )

    @property
    def request(self):
        return headers.header(self._req._headers)

    @request.setter
    def request(self, values_dict):
        self._req._headers.update(values_dict)
        if "Content-Type" in values_dict:
            self._req.ContentType = values_dict["Content-Type"]

    @property
    def all(self):
        return headers.header(self.request + self.response)


class cookies(object):
    class cookie(DotDict):
        def __str__(self):
            return "\n".join(["{}={}".format(k, v) for k, v in self.items()])

    def __init__(self, req):
        self._req = req

    @property
    def response(self):
        if self._req.response:
            c = self._req.response.getCookie().split("; ")
            if c[0]:
                return cookies.cookie(
                    {x[0]: x[2] for x in [x.partition("=") for x in c]}
                )

        return cookies.cookie({})

    @property
    def request(self):
        if "Cookie" in self._req._headers:
            c = self._req._headers["Cookie"].split("; ")
            if c[0]:
                return cookies.cookie(
                    {x[0]: x[2] for x in [x.partition("=") for x in c]}
                )

        return cookies.cookie({})

    @request.setter
    def request(self, values):
        self._req._headers["Cookie"] = "; ".join(values)

    @property
    def all(self):
        return cookies.cookie(self.request + self.response)


class params(object):
    class param(DotDict):
        def __str__(self):
            return "\n".join(["{}={}".format(k, v) for k, v in self.items()])

    def __init__(self, req):
        self._req = req

    @property
    def get(self):
        return params.param({x.name: x.value for x in self._req.getGETVars()})

    @get.setter
    def get(self, values):
        if isinstance(values, dict) or isinstance(values, DotDict):
            for key, value in values.items():
                self._req.setVariableGET(key, str(value))
        else:
            raise FuzzExceptBadAPI("GET Parameters must be specified as a dictionary")

    @property
    def post(self):
        return params.param({x.name: x.value for x in self._req.getPOSTVars()})

    @post.setter
    def post(self, pp):
        if isinstance(pp, dict) or isinstance(pp, DotDict):
            for key, value in pp.items():
                self._req.setVariablePOST(
                    key, str(value) if value is not None else value
                )

            self._req._non_parsed_post = self._req._variablesPOST.urlEncoded()

        elif isinstance(pp, str):
            self._req.setPostData(pp)

    @property
    def raw_post(self):
        return self._req._non_parsed_post

    @property
    def all(self):
        return params.param(self.get + self.post)

    @all.setter
    def all(self, values):
        self.get = values
        self.post = values


class FuzzRequest(FuzzRequestUrlMixing, FuzzRequestSoupMixing):
    def __init__(self):
        self._request = Request()

        self._proxy = None
        self._allvars = None
        self.wf_fuzz_methods = None
        self.wf_retries = 0
        self.wf_ip = None

        self.headers.request = {
            "User-Agent": Facade().sett.get("connection", "user-agent")
        }

    # methods for accessing HTTP requests information consistenly accross the codebase

    def __str__(self):
        return self._request.getAll()

    @property
    def raw_request(self):
        return self._request.getAll()

    @raw_request.setter
    def raw_request(self, rawReq, scheme):
        self.update_from_raw_http(rawReq, scheme)

    @property
    def raw_content(self):
        if self._request.response:
            return self._request.response.getAll()

        return ""

    @property
    def headers(self):
        return headers(self._request)

    @property
    def params(self):
        return params(self._request)

    @property
    def cookies(self):
        return cookies(self._request)

    @property
    def method(self):
        return self._request.method

    @method.setter
    def method(self, method):
        self._request.method = method

    @property
    def scheme(self):
        return self._request.schema

    @scheme.setter
    def scheme(self, s):
        self._request.schema = s

    @property
    def host(self):
        return self._request.getHost()

    @property
    def path(self):
        return self._request.path

    @property
    def redirect_url(self):
        return self._request.completeUrl

    @property
    def url(self):
        return self._request.finalUrl

    @url.setter
    def url(self, u):
        # urlparse goes wrong with IP:port without scheme (https://bugs.python.org/issue754016)
        if not u.startswith("FUZ") and (
            urlparse(u).netloc == "" or urlparse(u).scheme == ""
        ):
            u = "http://" + u

        if urlparse(u).path == "":
            u += "/"

        if Facade().sett.get("general", "encode_space") == "1":
            u = u.replace(" ", "%20")

        self._request.setUrl(u)
        if self.scheme.startswith("fuz") and self.scheme.endswith("z"):
            # avoid FUZZ to become fuzz
            self.scheme = self.scheme.upper()

    @property
    def content(self):
        return self._request.response.getContent() if self._request.response else ""

    @property
    def code(self):
        return self._request.response.code if self._request.response else 0

    @code.setter
    def code(self, c):
        self._request.response.code = int(c)

    @property
    def auth(self):
        method, creds = self._request.getAuth()

        return DotDict({"method": method, "credentials": creds})

    @auth.setter
    def auth(self, creds_dict):
        self._request.setAuth(creds_dict["method"], creds_dict["credentials"])
        method, creds = self._request.getAuth()

        return DotDict({"method": method, "credentials": creds})

    @property
    def follow(self):
        return self._request.followLocation

    @follow.setter
    def follow(self, f):
        self._request.setFollowLocation(f)

    @property
    def reqtime(self):
        return self._request.totaltime

    @reqtime.setter
    def reqtime(self, t):
        self._request.totaltime = t

    # Info extra that wfuzz needs within an HTTP request
    @property
    def wf_allvars_set(self):
        if self.wf_allvars == "allvars":
            return self.params.get
        elif self.wf_allvars == "allpost":
            return self.params.post
        elif self.wf_allvars == "allheaders":
            return self.headers.request
        else:
            raise FuzzExceptBadOptions("Unknown variable set: " + self.wf_allvars)

    @wf_allvars_set.setter
    def wf_allvars_set(self, varset):
        try:
            if self.wf_allvars == "allvars":
                self.params.get = varset
            elif self.wf_allvars == "allpost":
                self.params.post = varset
            elif self.wf_allvars == "allheaders":
                self._request.headers.request = varset
            else:
                raise FuzzExceptBadOptions("Unknown variable set: " + self.wf_allvars)
        except TypeError:
            raise FuzzExceptBadOptions(
                "It is not possible to use all fuzzing with duplicated parameters."
            )

    @property
    def wf_allvars(self):
        return self._allvars

    @wf_allvars.setter
    def wf_allvars(self, bl):
        if bl is not None and bl not in ["allvars", "allpost", "allheaders"]:
            raise FuzzExceptBadOptions(
                "Incorrect all parameters brute forcing type specified, correct values are allvars, allpost or allheaders."
            )

        self._allvars = bl

    @property
    def wf_proxy(self):
        return self._proxy

    @wf_proxy.setter
    def wf_proxy(self, proxy_tuple):
        if proxy_tuple:
            prox, ptype = proxy_tuple
            self._request.setProxy("%s" % prox, ptype if ptype else "HTML")
        self._proxy = proxy_tuple

    # methods wfuzz needs to perform HTTP requests (this might change in the future).

    def update_from_raw_http(self, raw, scheme, raw_response=None, raw_content=None):
        self._request.parseRequest(raw, scheme)

        # Parse request sets postdata = '' when there's POST request without data
        if self.method == "POST" and self.params.raw_post is None:
            self.params.post = ""

        if raw_response:
            rp = Response()
            if not isinstance(raw_response, str):
                raw_response = python2_3_convert_from_unicode(
                    raw_response.decode("utf-8", errors="surrogateescape")
                )
            rp.parseResponse(raw_response, raw_content)
            self._request.response = rp

        return self._request

    def to_cache_key(self):
        key = self._request.urlWithoutVariables

        dicc = {"g{}".format(key): True for key in self.params.get.keys()}
        dicc.update({"p{}".format(key): True for key in self.params.post.keys()})

        # take URL parameters into consideration
        url_params = list(dicc.keys())
        url_params.sort()
        key += "-" + "-".join(url_params)

        return key

    # methods wfuzz needs for substituing payloads and building dictionaries

    def update_from_options(self, options):
        if options["url"] != "FUZZ":
            self.url = options["url"]

        # headers must be parsed first as they might affect how reqresp parases other params
        self.headers.request = dict(options["headers"])

        if options["auth"].get("method") is not None:
            self.auth = options["auth"]

        if options["follow"]:
            self.follow = options["follow"]

        if options["postdata"] is not None:
            self.params.post = options["postdata"]

        if options["connect_to_ip"]:
            self.wf_ip = options["connect_to_ip"]

        if options["method"]:
            self.method = options["method"]
            self.wf_fuzz_methods = options["method"]

        if options["cookie"]:
            self.cookies.request = options["cookie"]

        if options["allvars"]:
            self.wf_allvars = options["allvars"]



import copy

from .fuzzfactory import reqfactory
from .payman import payman_factory

from ..fuzzobjects import FuzzResult, FuzzType, FuzzWord, FuzzWordType
from ..helpers.obj_factory import ObjectFactory, SeedBuilderHelper


class FuzzResultFactory(ObjectFactory):
    def __init__(self):
        ObjectFactory.__init__(
            self,
            {
                "fuzzres_from_options_and_dict": FuzzResultDictioBuilder(),
                "fuzzres_from_allvar": FuzzResultAllVarBuilder(),
                "fuzzres_from_recursion": FuzzResRecursiveBuilder(),
                "seed_from_recursion": SeedRecursiveBuilder(),
                "seed_from_options": SeedResultBuilder(),
                "seed_from_options_and_dict": FuzzResultDictSeedBuilder(),
                "baseline_from_options": BaselineResultBuilder(),
            },
        )


class FuzzResultDictioBuilder:
    def __call__(self, options, dictio_item):
        res = copy.deepcopy(options["compiled_seed"])
        res.item_type = FuzzType.RESULT
        res.discarded = False
        res.payload_man.update_from_dictio(dictio_item)
        res.update_from_options(options)

        SeedBuilderHelper.replace_markers(res.history, res.payload_man)
        res.nres = next(FuzzResult.newid)

        return res


class SeedResultBuilder:
    def __call__(self, options):
        seed = reqfactory.create("seed_from_options", options)
        res = FuzzResult(seed)
        res.payload_man = payman_factory.create("payloadman_from_request", seed)

        return res


class BaselineResultBuilder:
    def __call__(self, options):
        raw_seed = reqfactory.create("request_from_options", options)
        baseline_payloadman = payman_factory.create(
            "payloadman_from_baseline", raw_seed
        )

        if baseline_payloadman.payloads:
            res = FuzzResult(raw_seed)
            res.payload_man = baseline_payloadman
            res.update_from_options(options)
            res.is_baseline = True

            SeedBuilderHelper.replace_markers(raw_seed, baseline_payloadman)

            return res
        else:
            return None


class FuzzResultAllVarBuilder:
    def __call__(self, options, var_name, payload):
        fuzzres = copy.deepcopy(options["compiled_seed"])
        fuzzres.item_type = FuzzType.RESULT
        fuzzres.discarded = False
        fuzzres.payload_man = payman_factory.create("empty_payloadman", payload)
        fuzzres.payload_man.update_from_dictio([payload])
        fuzzres.history.wf_allvars_set = {var_name: payload.content}
        fuzzres.nres = next(FuzzResult.newid)

        return fuzzres


class FuzzResultDictSeedBuilder:
    def __call__(self, options, dictio):
        fuzzres = copy.deepcopy(dictio[0].content)
        fuzzres.history.update_from_options(options)
        fuzzres.update_from_options(options)
        fuzzres.payload_man = payman_factory.create("empty_payloadman", dictio[0])
        fuzzres.payload_man.update_from_dictio(dictio)

        return fuzzres


class SeedRecursiveBuilder:
    def __call__(self, seed):
        new_seed = copy.deepcopy(seed)
        new_seed.history.url = seed.history.recursive_url + "FUZZ"
        new_seed.rlevel += 1
        if new_seed.rlevel_desc:
            new_seed.rlevel_desc += " - "
        new_seed.rlevel_desc += seed.payload_man.description()
        new_seed.item_type = FuzzType.SEED
        new_seed.discarded = False
        new_seed.payload_man = payman_factory.create(
            "payloadman_from_request", new_seed.history
        )

        return new_seed


class FuzzResRecursiveBuilder:
    def __call__(self, seed, url):
        fr = copy.deepcopy(seed)
        fr.history.url = str(url)
        fr.rlevel = seed.rlevel + 1
        if fr.rlevel_desc:
            fr.rlevel_desc += " - "
        fr.rlevel_desc += seed.payload_man.description()
        fr.item_type = FuzzType.BACKFEED
        fr.discarded = False
        fr.is_baseline = False

        fr.payload_man = payman_factory.create(
            "empty_payloadman", FuzzWord(url, FuzzWordType.WORD)
        )

        return fr


resfactory = FuzzResultFactory()



from wfuzz.plugin_api.base import BasePlugin
from wfuzz.externals.moduleman.plugin import moduleman_plugin

import re

KBASE_KEY = "http.servers"
KBASE_KEY_RESP_UNCOMMON = "http.response.headers.uncommon"
KBASE_KEY_REQ_UNCOMMON = "http.request.headers.uncommon"

SERVER_HEADERS = ["server", "x-powered-by" "via"]

COMMON_RESPONSE_HEADERS_REGEX_LIST = [
    r"^Server$",
    r"^X-Powered-By$",
    r"^Via$",
    r"^Access-Control.*$",
    r"^Accept-.*$",
    r"^age$",
    r"^allow$",
    r"^Cache-control$",
    r"^Client-.*$",
    r"^Connection$",
    r"^Content-.*$",
    r"^Cross-Origin-Resource-Policy$",
    r"^Date$",
    r"^Etag$",
    r"^Expires$",
    r"^Keep-Alive$",
    r"^Last-Modified$",
    r"^Link$",
    r"^Location$",
    r"^P3P$",
    r"^Pragma$",
    r"^Proxy-.*$",
    r"^Refresh$",
    r"^Retry-After$",
    r"^Referrer-Policy$",
    r"^Set-Cookie$",
    r"^Server-Timing$",
    r"^Status$",
    r"^Strict-Transport-Security$",
    r"^Timing-Allow-Origin$",
    r"^Trailer$",
    r"^Transfer-Encoding$",
    r"^Upgrade$",
    r"^Vary$",
    r"^Warning^$",
    r"^WWW-Authenticate$",
    r"^X-Content-Type-Options$",
    r"^X-Download-Options$",
    r"^X-Frame-Options$",
    r"^X-Microsite$",
    r"^X-Request-Handler-Origin-Region$",
    r"^X-XSS-Protection$",
]

COMMON_RESPONSE_HEADERS_REGEX = re.compile(
    "({})".format("|".join(COMMON_RESPONSE_HEADERS_REGEX_LIST)), re.IGNORECASE
)

COMMON_REQ_HEADERS_REGEX_LIST = [
    r"A-IM$",
    r"Accept$",
    r"Accept-.*$",
    r"Access-Control-.*$",
    r"Authorization$",
    r"Cache-Control$",
    r"Connection$",
    r"Content-.*$",
    r"Cookie$",
    r"Date$",
    r"Expect$",
    r"Forwarded$",
    r"From$",
    r"Host$",
    r"If-.*$",
    r"Max-Forwards$",
    r"Origin$",
    r"Pragma$",
    r"Proxy-Authorization$",
    r"Range$",
    r"Referer$",
    r"TE$",
    r"User-Agent$",
    r"Upgrade$",
    r"Upgrade-Insecure-Requests$",
    r"Via$",
    r"Warning$",
    r"X-Requested-With$",
    r"X-HTTP-Method-Override$",
    r"X-Requested-With$",
]

COMMON_REQ_HEADERS_REGEX = re.compile(
    "({})".format("|".join(COMMON_REQ_HEADERS_REGEX_LIST)), re.IGNORECASE
)


@moduleman_plugin
class headers(BasePlugin):
    name = "headers"
    author = ("Xavi Mendez (@xmendez)",)
    version = "0.1"
    summary = "Looks for HTTP headers."
    description = (
        "Looks for NEW HTTP headers:",
        "\t- Response HTTP headers associated to web servers.",
        "\t- Uncommon response HTTP headers.",
        "\t- Uncommon request HTTP headers.",
        "It is worth noting that, only the FIRST match of the above headers is registered.",
    )
    category = ["info", "passive", "default"]
    priority = 99
    parameters = ()

    def __init__(self):
        BasePlugin.__init__(self)

    def validate(self, fuzzresult):
        return True

    def check_request_header(self, fuzzresult, header, value):
        header_value = None
        if not COMMON_REQ_HEADERS_REGEX.match(header):
            header_value = header

        if header_value is not None:
            if (
                header_value.lower() not in self.kbase[KBASE_KEY_REQ_UNCOMMON]
                or KBASE_KEY_REQ_UNCOMMON not in self.kbase
            ):
                self.add_result(
                    "reqheader",
                    "New uncommon HTTP request header",
                    "{}: {}".format(header_value, value),
                )

                self.kbase[KBASE_KEY_REQ_UNCOMMON].append(header_value.lower())

    def check_response_header(self, fuzzresult, header, value):
        header_value = None
        if not COMMON_RESPONSE_HEADERS_REGEX.match(header):
            header_value = header

        if header_value is not None:
            if (
                header_value.lower() not in self.kbase[KBASE_KEY_RESP_UNCOMMON]
                or KBASE_KEY_RESP_UNCOMMON not in self.kbase
            ):
                self.add_result(
                    "header",
                    "New uncommon HTTP response header",
                    "{}: {}".format(
                        header_value, fuzzresult.history.headers.response[header_value],
                    ),
                )

                self.kbase[KBASE_KEY_RESP_UNCOMMON].append(header_value.lower())

    def check_server_header(self, fuzzresult, header, value):
        if header.lower() in SERVER_HEADERS:
            if (
                value.lower() not in self.kbase[KBASE_KEY]
                or KBASE_KEY not in self.kbase
            ):
                self.add_result(
                    "server", "New server HTTP response header", "{}".format(value),
                )

                self.kbase[KBASE_KEY].append(value.lower())

    def process(self, fuzzresult):
        for header, value in fuzzresult.history.headers.request.items():
            self.check_request_header(fuzzresult, header, value)

        for header, value in fuzzresult.history.headers.response.items():
            self.check_response_header(fuzzresult, header, value)
            self.check_server_header(fuzzresult, header, value)



from .options import FuzzSession
from .facade import Facade
from .ui.console.clparser import CLParser

"""
Wfuzz API
"""


def fuzz(**kwargs):
    return FuzzSession(**kwargs).fuzz()


def get_payloads(iterator):
    fs = FuzzSession()

    return fs.get_payloads(iterator)


def get_payload(iterator):
    fs = FuzzSession()
    return fs.get_payload(iterator)


def encode(name, value):
    return Facade().encoders.get_plugin(name)().encode(value)


def decode(name, value):
    return Facade().encoders.get_plugin(name)().decode(value)


def payload(**kwargs):
    return FuzzSession(**kwargs).payload()


def get_session(cline):
    cl = ["wfuzz"] + cline.split(" ")
    return FuzzSession(**CLParser(cl).parse_cl())



from wfuzz.externals.moduleman.plugin import moduleman_plugin
from wfuzz.plugin_api.base import BasePayload
from wfuzz.exception import FuzzExceptBadOptions
from wfuzz.fuzzobjects import FuzzWordType


@moduleman_plugin
class hexrange(BasePayload):
    name = "hexrange"
    author = (
        "Carlos del Ojo",
        "Christian Martorella",
        "Adapted to newer versions Xavi Mendez (@xmendez)",
    )
    version = "0.1"
    description = ()
    summary = "Returns each hex number of the given hex range."
    category = ["default"]
    priority = 99

    parameters = (
        ("range", "", True, "Range of hex numbers to generate in the form of 00-ff."),
    )

    default_parameter = "range"

    def __init__(self, params):
        BasePayload.__init__(self, params)

        try:
            ran = self.params["range"].split("-")
            self.minimum = int(ran[0], 16)
            self.maximum = int(ran[1], 16)
            self.__count = self.maximum - self.minimum + 1
            self.current = self.minimum
            self.lgth = max(
                len(ran[0]), len(ran[1]), len(hex(self.maximum).replace("0x", ""))
            )
        except ValueError:
            raise FuzzExceptBadOptions('Bad range format (eg. "0-ffa")')

    def count(self):
        return self.__count

    def get_type(self):
        return FuzzWordType.WORD

    def get_next(self):
        if self.current > self.maximum:
            raise StopIteration

        pl = "%" + str(self.lgth) + "s"
        num = hex(self.current).replace("0x", "")
        pl = pl % (num)
        payl = pl.replace(" ", "0")

        self.current += 1

        return payl



from wfuzz.plugin_api.base import BasePlugin
from wfuzz.externals.moduleman.plugin import moduleman_plugin

import subprocess
import tempfile
import pipes
import os
import re


@moduleman_plugin
class screenshot(BasePlugin):
    name = "screenshot"
    author = ("Xavi Mendez (@xmendez)",)
    version = "0.1"
    summary = "Performs a screen capture using linux cutycapt tool"
    description = (
        "Performs a screen capture using linux cutycapt tool",
        "The tool must be installed and in the executable path",
    )
    category = ["tools", "active"]
    priority = 99

    parameters = ()

    def __init__(self):
        BasePlugin.__init__(self)

    def validate(self, fuzzresult):
        return fuzzresult.code not in [404]

    def process(self, fuzzresult):
        temp_name = next(tempfile._get_candidate_names())
        defult_tmp_dir = tempfile._get_default_tempdir()

        filename = os.path.join(
            defult_tmp_dir,
            (temp_name + "_" + re.sub(r"[^a-zA-Z0-9_-]", "_", fuzzresult.url))[:200]
            + ".jpg",
        )

        subprocess.call(
            [
                "cutycapt",
                "--url=%s" % pipes.quote(fuzzresult.url),
                "--out=%s" % filename,
                "--insecure",
                "--print-backgrounds=on",
            ]
        )
        self.add_result("file", "Screnshot taken", filename)



from .plugin_api.urlutils import parse_url
from .exception import FuzzExceptBadInstall

# python 2 and 3
import sys

if sys.version_info >= (3, 0):
    from urllib.parse import urljoin, urlparse
else:
    from urlparse import urljoin, urlparse


class FuzzRequestSoupMixing(object):
    def get_soup(self):
        try:
            from bs4 import BeautifulSoup
        except ImportError:
            raise FuzzExceptBadInstall("You need to install beautifulsoup4 first!")

        soup = BeautifulSoup(self.content, "html.parser")

        return soup


class FuzzRequestUrlMixing(object):
    # urlparse functions
    @property
    def urlparse(self):
        return parse_url(self.url)

    @property
    def urlp(self):
        return parse_url(self.url)

    @property
    def pstrip(self):
        return self.to_cache_key()

    @property
    def is_path(self):
        if self.recursive_url and self.recursive_url[-1] == "/":
            return True

        return False

    @property
    def recursive_url(self):
        if self.code >= 300 and self.code < 308 and "Location" in self.headers.response:
            location_url = self.headers.response["Location"]
            location_parsed_url = urlparse(location_url)

            if not location_parsed_url.scheme and not location_parsed_url.netloc:
                return urljoin(self.url, location_url)
        elif self.code in [200, 401] and self.url[-1] == "/":
            return self.url

        return None



from wfuzz.externals.moduleman.plugin import moduleman_plugin
from wfuzz.plugin_api.base import BasePayload
from wfuzz.fuzzobjects import FuzzWordType

from wfuzz.facade import Facade


@moduleman_plugin
class guitab(BasePayload):
    name = "guitab"
    author = ("Xavi Mendez (@xmendez)",)
    version = "0.1"
    description = (
        "** This is a beta plugin for the GUI under construction.",
        "This payload reads requests from a tab in the GUI",
    )
    summary = "This payload reads requests from a tab in the GUI"
    category = ["default"]
    priority = 99

    parameters = (
        ("tab", "", True, "Name of a valid GUI tab."),
        (
            "attr",
            None,
            False,
            "Attribute of fuzzresult to return. If not specified the whole object is returned.",
        ),
    )

    default_parameter = "tab"

    def __init__(self, params):
        BasePayload.__init__(self, params)

        self.attr = self.params["attr"]
        self._it = iter(Facade().data[self.params["tab"]])

    def count(self):
        return len(Facade().data[self.params["tab"]])

    def get_type(self):
        return FuzzWordType.WORD

    def get_next(self):
        next_item = next(self._it)

        return next_item if not self.attr else next_item.get_field(self.attr)



from ..fuzzrequest import FuzzRequest

from ..helpers.obj_factory import ObjectFactory, SeedBuilderHelper


class FuzzRequestFactory(ObjectFactory):
    def __init__(self):
        ObjectFactory.__init__(
            self,
            {
                "request_from_options": RequestBuilder(),
                "seed_from_options": SeedBuilder(),
            },
        )


class RequestBuilder:
    def __call__(self, options):
        fr = FuzzRequest()

        fr.url = options["url"]
        fr.wf_fuzz_methods = options["method"]
        fr.update_from_options(options)

        return fr


class SeedBuilder:
    def __call__(self, options):
        seed = reqfactory.create("request_from_options", options)
        marker_dict = SeedBuilderHelper.get_marker_dict(seed)
        SeedBuilderHelper.remove_baseline_markers(seed, marker_dict)

        return seed


reqfactory = FuzzRequestFactory()



from threading import Lock
import difflib


class MyCounter:
    def __init__(self, count=0):
        self._count = count
        self._mutex = Lock()

    def inc(self):
        return self._operation(1)

    def dec(self):
        return self._operation(-1)

    def _operation(self, dec):
        with self._mutex:
            self._count += dec
            return self._count

    def __call__(self):
        with self._mutex:
            return self._count


def diff(param1, param2):
    delta = difflib.unified_diff(
        str(param1).splitlines(False),
        str(param2).splitlines(False),
        fromfile="prev",
        tofile="current",
        n=0,
    )

    return "\n".join(delta)



__title__ = "wfuzz"
__version__ = "3.1.0"
__build__ = 0x023000
__author__ = "Xavier Mendez"
__license__ = "GPL 2.0"
__copyright__ = "Copyright 2011-2020 Xavier Mendez"

import logging
import sys

import warnings


# define a logging Handler
console = logging.StreamHandler()
console.setLevel(logging.WARNING)
formatter = logging.Formatter("%(name)-12s: %(levelname)-8s %(message)s")
console.setFormatter(formatter)
logging.getLogger("").addHandler(console)


# define warnings format
def warning_on_one_line(message, category, filename, lineno, file=None, line=None):
    return " %s:%s: %s:%s\n" % (filename, lineno, category.__name__, message)


warnings.formatwarning = warning_on_one_line


try:
    import pycurl

    if "libressl".lower() not in pycurl.version.lower():
        warnings.warn(
            "Pycurl is not compiled against Openssl. Wfuzz might not work correctly when fuzzing SSL sites. Check Wfuzz's documentation for more information."
        )

    if not hasattr(pycurl, "CONNECT_TO"):
        warnings.warn(
            "Pycurl and/or libcurl version is old. CONNECT_TO option is missing. Wfuzz --ip option will not be available."
        )

    if not hasattr(pycurl, "PATH_AS_IS"):
        warnings.warn(
            "Pycurl and/or libcurl version is old. PATH_AS_IS option is missing. Wfuzz might not correctly fuzz URLS with '..'."
        )

except ImportError:
    warnings.warn(
        "fuzz needs pycurl to run. Pycurl could be installed using the following command: $ pip install pycurl"
    )

    sys.exit(1)

from .options import FuzzSession
from .api import fuzz, get_payload, get_payloads, encode, decode, payload, get_session



from wfuzz.externals.moduleman.plugin import moduleman_plugin
from wfuzz.plugin_api.base import BasePlugin

# Python 2 and 3
try:
    from urllib.parse import urljoin
except ImportError:
    from urlparse import urljoin


@moduleman_plugin
class backups(BasePlugin):
    name = "backups"
    summary = "Looks for known backup filenames."
    description = ("Looks for known backup filenames.",)
    "For example, given http://localhost.com/dir/index.html, it will perform the following requests",
    "* http://localhost/dir/index.EXTENSIONS",
    "* http://localhost/dir/index.html.EXTENSIONS",
    "* http://localhost/dir.EXTENSIONS",
    author = ("Xavi Mendez (@xmendez)",)
    version = "0.1"
    category = ["fuzzer", "active"]
    priority = 99

    parameters = (
        (
            "ext",
            ".bak,.tgz,.zip,.tar.gz,~,.rar,.old,.-.swp",
            False,
            "Extensions to look for.",
        ),
    )

    def __init__(self):
        BasePlugin.__init__(self)
        self.extensions = self.kbase["backups.ext"][0].split(",")

    def validate(self, fuzzresult):
        return fuzzresult.code != 404 and (
            fuzzresult.history.urlparse.fext not in self.extensions
        )

    def process(self, fuzzresult):
        # >>> urlparse.urlparse("http://www.localhost.com/kk/index.html?id=1")
        # ParseResult(scheme='http', netloc='www.localhost.com', path='/kk/index.html', params='', query='id=1', fragment='')

        for pre_extension in self.extensions:
            pre, nothing, extension = pre_extension.partition("-")

            # http://localhost/dir/test.html -----> test.BAKKK
            self.queue_url(
                urljoin(
                    fuzzresult.url, pre + fuzzresult.history.urlparse.fname + extension
                )
            )

            # http://localhost/dir/test.html ---> test.html.BAKKK
            self.queue_url(
                urljoin(fuzzresult.url, fuzzresult.history.urlparse.ffname + extension)
            )

            # http://localhost/dir/test.html ----> dir.BAKKK



import abc

import pycurl

from ..helpers.obj_factory import HttpRequestFactory
from ..helpers.str_func import (
    python2_3_convert_to_unicode,
    python2_3_convert_from_unicode,
)


from ..externals.reqresp import Response


PYCURL_PATH_AS_IS = True
if not hasattr(pycurl, "PATH_AS_IS"):
    PYCURL_PATH_AS_IS = False


class ReqRespRequestFactory(HttpRequestFactory):
    def to_http_object(options, req, pycurl_c):
        pycurl_c.setopt(pycurl.MAXREDIRS, 5)

        pycurl_c.setopt(pycurl.WRITEFUNCTION, req._request.body_callback)
        pycurl_c.setopt(pycurl.HEADERFUNCTION, req._request.header_callback)

        pycurl_c.setopt(pycurl.NOSIGNAL, 1)
        pycurl_c.setopt(pycurl.SSL_VERIFYPEER, False)
        pycurl_c.setopt(pycurl.SSL_VERIFYHOST, 0)

        if PYCURL_PATH_AS_IS:
            pycurl_c.setopt(pycurl.PATH_AS_IS, 1)

        pycurl_c.setopt(
            pycurl.URL, python2_3_convert_to_unicode(req._request.completeUrl)
        )

        if req._request.getConnTimeout():
            pycurl_c.setopt(pycurl.CONNECTTIMEOUT, req._request.getConnTimeout())

        if req._request.getTotalTimeout():
            pycurl_c.setopt(pycurl.TIMEOUT, req._request.getTotalTimeout())

        authMethod, userpass = req._request.getAuth()
        if authMethod or userpass:
            if authMethod == "basic":
                pycurl_c.setopt(pycurl.HTTPAUTH, pycurl.HTTPAUTH_BASIC)
            elif authMethod == "ntlm":
                pycurl_c.setopt(pycurl.HTTPAUTH, pycurl.HTTPAUTH_NTLM)
            elif authMethod == "digest":
                pycurl_c.setopt(pycurl.HTTPAUTH, pycurl.HTTPAUTH_DIGEST)
            pycurl_c.setopt(pycurl.USERPWD, python2_3_convert_to_unicode(userpass))
        else:
            pycurl_c.unsetopt(pycurl.USERPWD)

        pycurl_c.setopt(
            pycurl.HTTPHEADER, python2_3_convert_to_unicode(req._request.getHeaders())
        )

        curl_options = {
            "GET": pycurl.HTTPGET,
            "POST": pycurl.POST,
            "PATCH": pycurl.UPLOAD,
            "HEAD": pycurl.NOBODY,
        }

        for verb in curl_options.values():
            pycurl_c.setopt(verb, False)

        if req._request.method in curl_options:
            pycurl_c.unsetopt(pycurl.CUSTOMREQUEST)
            pycurl_c.setopt(curl_options[req._request.method], True)
        else:
            pycurl_c.setopt(pycurl.CUSTOMREQUEST, req._request.method)

        if req._request._non_parsed_post is not None:
            pycurl_c.setopt(
                pycurl.POSTFIELDS,
                python2_3_convert_to_unicode(req._request._non_parsed_post),
            )

        pycurl_c.setopt(pycurl.FOLLOWLOCATION, 1 if req._request.followLocation else 0)

        # proxy = req._request.getProxy()
        # if proxy is not None:
        #     pycurl_c.setopt(pycurl.PROXY, python2_3_convert_to_unicode(proxy))
        #     if req._request.proxytype == "SOCKS5":
        #         pycurl_c.setopt(pycurl.PROXYTYPE, pycurl.PROXYTYPE_SOCKS5)
        #     elif req._request.proxytype == "SOCKS4":
        #         pycurl_c.setopt(pycurl.PROXYTYPE, pycurl.PROXYTYPE_SOCKS4)
        #     req._request.delHeader("Proxy-Connection")
        # else:
        #     pycurl_c.setopt(pycurl.PROXY, "")

        if req.wf_ip:
            pycurl_c.setopt(
                pycurl.CONNECT_TO,
                ["::{}:{}".format(req.wf_ip["ip"], req.wf_ip["port"])],
            )

        return pycurl_c

    def from_http_object(options, req, pycurl_c, header, body):
        raw_header = python2_3_convert_from_unicode(
            header.decode("utf-8", errors="surrogateescape")
        )

        if pycurl_c.getinfo(pycurl.EFFECTIVE_URL) != req._request.completeUrl:
            req._request.setFinalUrl(pycurl_c.getinfo(pycurl.EFFECTIVE_URL))

        req._request.totaltime = pycurl_c.getinfo(pycurl.TOTAL_TIME)

        req._request.response = Response()
        req._request.response.parseResponse(raw_header, rawbody=body)

        return req._request.response



from wfuzz.externals.moduleman.plugin import moduleman_plugin
from wfuzz.plugin_api.base import BasePayload
from wfuzz.fuzzobjects import FuzzWordType


@moduleman_plugin
class list(BasePayload):
    name = "list"
    author = ("Xavi Mendez (@xmendez)",)
    version = "0.1"
    description = ("ie word1-word2",)
    summary = "Returns each element of the given word list separated by -."
    category = ["default"]
    priority = 99

    parameters = (
        ("values", "", True, "Values separated by - to return as a dictionary."),
    )

    default_parameter = "values"

    def __init__(self, params):
        BasePayload.__init__(self, params)

        if self.params["values"].find("\\") >= 0:
            self.params["values"] = self.params["values"].replace("\\-", "$SEP$")
            self.params["values"] = self.params["values"].replace("\\\\", "$SCAP$")

            self.value_list = self.params["values"].split("-")

            for i in range(len(self.value_list)):
                self.value_list[i] = self.value_list[i].replace("$SEP$", "-")
                self.value_list[i] = self.value_list[i].replace("$SCAP$", "\\")
        else:
            self.value_list = self.params["values"].split("-")

        self.__count = len(self.value_list)
        self.current = 0

    def count(self):
        return self.__count

    def get_type(self):
        return FuzzWordType.WORD

    def get_next(self):
        if self.current >= self.__count:
            raise StopIteration
        else:
            elem = self.value_list[self.current]
            self.current += 1
            return elem



from wfuzz.externals.moduleman.plugin import moduleman_plugin
from wfuzz.plugin_api.base import BasePayload
from wfuzz.exception import FuzzExceptPluginBadParams
from wfuzz.fuzzobjects import FuzzWordType

import random


@moduleman_plugin
class hexrand(BasePayload):
    name = "hexrand"
    author = (
        "Carlos del Ojo",
        "Christian Martorella",
        "Adapted to newer versions Xavi Mendez (@xmendez)",
    )
    version = "0.1"
    description = ()
    summary = "Returns random hex numbers from the given range."
    category = ["default"]
    priority = 99

    parameters = (
        (
            "range",
            "",
            True,
            "Range of hex numbers to randomly generate in the form of 00-ff.",
        ),
    )

    default_parameter = "range"

    def __init__(self, params):
        BasePayload.__init__(self, params)

        try:
            ran = self.params["range"].split("-")
            self.minimum = int(ran[0], 16)
            self.maximum = int(ran[1], 16)
            self.__count = -1
        except ValueError:
            raise FuzzExceptPluginBadParams('Bad range format (eg. "0-ffa")')

    def __iter__(self):
        return self

    def count(self):
        return self.__count

    def get_type(self):
        return FuzzWordType.WORD

    def get_next(self):
        self.current = random.SystemRandom().randint(self.minimum, self.maximum)

        lgth = len(hex(self.maximum).replace("0x", ""))
        pl = "%" + str(lgth) + "s"
        num = hex(self.current).replace("0x", "")
        pl = pl % (num)
        payl = pl.replace(" ", "0")

        return payl



from .exception import (
    FuzzExceptBadRecipe,
    FuzzExceptBadOptions,
    FuzzExceptBadFile,
)
from .facade import (
    Facade,
    ERROR_CODE,
    BASELINE_CODE,
)

from .factories.fuzzresfactory import resfactory
from .factories.dictfactory import dictionary_factory
from .fuzzobjects import FuzzStats
from .filters.ppfilter import FuzzResFilter
from .filters.simplefilter import FuzzResSimpleFilter
from .helpers.str_func import (
    json_minify,
    python2_3_convert_from_unicode,
)

from .core import Fuzzer
from .myhttp import HttpPool

from .externals.reqresp.cache import HttpCache

from collections import defaultdict

# python 2 and 3
try:
    from collections import UserDict
except ImportError:
    from UserDict import UserDict

import json


class FuzzSession(UserDict):
    def __init__(self, **kwargs):
        self.data = self._defaults()
        self.keys_not_to_dump = [
            "interactive",
            "recipe",
            "seed_payload",
            "compiled_stats",
            "compiled_dictio",
            "compiled_simple_filter",
            "compiled_filter",
            "compiled_prefilter",
            "compiled_printer",
            "description",
            "show_field",
            "transport",
        ]

        # recipe must be superseded by options
        if "recipe" in kwargs and kwargs["recipe"]:
            for recipe in kwargs["recipe"]:
                self.import_from_file(recipe)

        self.update(kwargs)

        self.cache = HttpCache()
        self.http_pool = None

        self.stats = FuzzStats()

    def _defaults(self):
        return dict(
            console_printer="",
            hs=None,
            hc=[],
            hw=[],
            hl=[],
            hh=[],
            ss=None,
            sc=[],
            sw=[],
            sl=[],
            sh=[],
            payloads=None,
            iterator=None,
            printer=(None, None),
            colour=False,
            previous=False,
            verbose=False,
            interactive=False,
            transport="http",
            recipe=[],
            save="",
            proxies=None,
            conn_delay=int(Facade().sett.get("connection", "conn_delay")),
            req_delay=int(Facade().sett.get("connection", "req_delay")),
            retries=int(Facade().sett.get("connection", "retries")),
            rlevel=0,
            dlevel=4,
            scanmode=False,
            delay=None,
            concurrent=int(Facade().sett.get("connection", "concurrent")),
            url="",
            method=None,
            auth={},
            follow=False,
            postdata=None,
            headers=[],
            cookie=[],
            allvars=None,
            script="",
            script_args={},
            connect_to_ip=None,
            fields=[],
            no_cache=False,
            show_field=None,
            # this is equivalent to payloads but in a different format
            dictio=None,
            # these will be compiled
            seed_payload=False,
            filter="",
            prefilter=[],
            compiled_filter=None,
            compiled_prefilter=[],
            compiled_printer=None,
            compiled_seed=None,
            compiled_baseline=None,
            compiled_stats=None,
            compiled_dictio=None,
            exec_mode="api",
        )

    def update(self, options):
        self.data.update(options)

    def validate(self):
        error_list = []

        if self.data["dictio"] and self.data["payloads"]:
            raise FuzzExceptBadOptions(
                "Bad usage: Dictio and payloads options are mutually exclusive. Only one could be specified."
            )

        if self.data["rlevel"] > 0 and self.data["transport"] == "dryrun":
            error_list.append(
                "Bad usage: Recursion cannot work without making any HTTP request."
            )

        if self.data["script"] and self.data["transport"] == "dryrun":
            error_list.append(
                "Bad usage: Plugins cannot work without making any HTTP request."
            )

        if self.data["no_cache"] not in [True, False]:
            raise FuzzExceptBadOptions("Bad usage: No-cache is a boolean value")

        if not self.data["url"]:
            error_list.append("Bad usage: You must specify an URL.")

        if not self.data["payloads"] and not self.data["dictio"]:
            error_list.append("Bad usage: You must specify a payload.")

        if self.data["hs"] and self.data["ss"]:
            raise FuzzExceptBadOptions(
                "Bad usage: Hide and show regex filters flags are mutually exclusive. Only one could be specified."
            )

        if self.data["rlevel"] < 0:
            raise FuzzExceptBadOptions(
                "Bad usage: Recursion level must be a positive int."
            )

        if self.data["allvars"] not in [None, "allvars", "allpost", "allheaders"]:
            raise FuzzExceptBadOptions(
                "Bad options: Incorrect all parameters brute forcing type specified, correct values are allvars,allpost or allheaders."
            )

        if self.data["proxies"]:
            for ip, port, ttype in self.data["proxies"]:
                if ttype not in ("SOCKS5", "SOCKS4", "HTTP"):
                    raise FuzzExceptBadOptions(
                        "Bad proxy type specified, correct values are HTTP, SOCKS4 or SOCKS5."
                    )

        return error_list

    def export_to_file(self, filename):
        try:
            with open(filename, "w") as f:
                f.write(self.export_json())
        except IOError:
            raise FuzzExceptBadFile("Error writing recipe file.")

    def import_from_file(self, filename):
        try:
            with open(filename, "r") as f:
                self.import_json(f.read())
        except IOError:
            raise FuzzExceptBadFile("Error loading recipe file {}.".format(filename))
        except json.decoder.JSONDecodeError as e:
            raise FuzzExceptBadRecipe(
                "Incorrect JSON recipe {} format: {}".format(filename, str(e))
            )

    def import_json(self, data):
        js = json.loads(json_minify(data))

        try:
            if js["version"] == "0.2" and "wfuzz_recipe" in js:
                for k, v in js["wfuzz_recipe"].items():
                    if k not in self.keys_not_to_dump:
                        # python 2 and 3 hack
                        if k in self.data and isinstance(self.data[k], list):
                            self.data[k] += python2_3_convert_from_unicode(v)
                        else:
                            self.data[k] = python2_3_convert_from_unicode(v)
            else:
                raise FuzzExceptBadRecipe("Unsupported recipe version.")
        except KeyError:
            raise FuzzExceptBadRecipe("Incorrect recipe format.")

    def export_json(self):
        tmp = dict(version="0.2", wfuzz_recipe=defaultdict(dict))
        defaults = self._defaults()

        # Only dump the non-default options
        for k, v in self.data.items():
            if v != defaults[k] and k not in self.keys_not_to_dump:
                tmp["wfuzz_recipe"][k] = self.data[k]

        return json.dumps(tmp, sort_keys=True, indent=4, separators=(",", ": "))

    def payload(self, **kwargs):
        try:
            self.data.update(kwargs)
            self.compile_seeds()
            self.compile_dictio()
            for r in self.data["compiled_dictio"]:
                yield tuple((fuzz_word.content for fuzz_word in r))
        finally:
            self.data["compiled_dictio"].cleanup()

    def fuzz(self, **kwargs):
        self.data.update(kwargs)

        fz = None
        try:
            fz = Fuzzer(self.compile())

            for f in fz:
                yield f

        finally:
            if fz:
                fz.cancel_job()
                self.stats.update(self.data["compiled_stats"])

            if self.http_pool:
                self.http_pool.deregister()
                self.http_pool = None

    def get_payloads(self, iterator):
        self.data["dictio"] = iterator

        return self

    def get_payload(self, iterator):
        return self.get_payloads([iterator])

    def __enter__(self):
        self.http_pool = HttpPool(self)
        self.http_pool.register()
        return self

    def __exit__(self, *args):
        self.close()

    def get_fuzz_words(self):
        fuzz_words = self.data["compiled_filter"].get_fuzz_words()

        for comp_obj in ["compiled_seed", "compiled_baseline"]:
            if self.data[comp_obj]:
                fuzz_words += self.data[comp_obj].payload_man.get_fuzz_words()

        for prefilter in self.data["compiled_prefilter"]:
            fuzz_words += prefilter.get_fuzz_words()

        if self.data["url"] == "FUZZ":
            fuzz_words.append("FUZZ")

        return set(fuzz_words)

    def compile_dictio(self):
        if self.data["allvars"]:
            self.data["compiled_dictio"] = dictionary_factory.create(
                "dictio_from_allvar", self
            )
        else:
            self.data["compiled_dictio"] = dictionary_factory.create(
                "dictio_from_options", self
            )

    def compile_seeds(self):
        self.data["compiled_seed"] = resfactory.create("seed_from_options", self)
        self.data["compiled_baseline"] = resfactory.create(
            "baseline_from_options", self
        )

    def compile(self):
        # Validate options
        error = self.validate()
        if error:
            raise FuzzExceptBadOptions(error[0])

        self.data["seed_payload"] = True if self.data["url"] == "FUZZ" else False

        # printer
        try:
            filename, printer = self.data["printer"]
        except ValueError:
            raise FuzzExceptBadOptions(
                "Bad options: Printer must be specified in the form of ('filename', 'printer')"
            )

        if filename:
            if printer == "default" or not printer:
                printer = Facade().sett.get("general", "default_printer")
            self.data["compiled_printer"] = Facade().printers.get_plugin(printer)(
                filename
            )

        try:
            for filter_option in ["hc", "hw", "hl", "hh", "sc", "sw", "sl", "sh"]:
                self.data[filter_option] = [
                    BASELINE_CODE
                    if i == "BBB"
                    else ERROR_CODE
                    if i == "XXX"
                    else int(i)
                    for i in self.data[filter_option]
                ]
        except ValueError:
            raise FuzzExceptBadOptions(
                "Bad options: Filter must be specified in the form of [int, ... , int, BBB, XXX]."
            )

        self.compile_seeds()
        self.compile_dictio()

        # filter options
        self.data["compiled_simple_filter"] = FuzzResSimpleFilter.from_options(self)
        self.data["compiled_filter"] = FuzzResFilter(self.data["filter"])
        for prefilter in self.data["prefilter"]:
            self.data["compiled_prefilter"].append(
                FuzzResFilter(filter_string=prefilter)
            )

        self.data["compiled_stats"] = FuzzStats.from_options(self)

        # Check payload num
        fuzz_words = self.get_fuzz_words()

        if self.data["compiled_dictio"].width() != len(fuzz_words):
            raise FuzzExceptBadOptions(
                "FUZZ words and number of payloads do not match!"
            )

        if self.data["allvars"] is None and len(fuzz_words) == 0:
            raise FuzzExceptBadOptions("You must specify at least a FUZZ word!")

        if self.data["compiled_baseline"] is None and (
            BASELINE_CODE in self.data["hc"]
            or BASELINE_CODE in self.data["hl"]
            or BASELINE_CODE in self.data["hw"]
            or BASELINE_CODE in self.data["hh"]
        ):
            raise FuzzExceptBadOptions(
                "Bad options: specify a baseline value when using BBB"
            )

        if self.data["script"]:
            Facade().scripts.kbase.update(self.data["script_args"])

            for k, v in Facade().sett.get_section("kbase"):
                if k not in self.data["script_args"]:
                    Facade().scripts.kbase[k] = v

        if not self.http_pool:
            self.http_pool = HttpPool(self)
            self.http_pool.register()

        if self.data["colour"]:
            Facade().printers.kbase["colour"] = True

        if self.data["verbose"]:
            Facade().printers.kbase["verbose"] = True

        return self

    def close(self):
        if self.data["compiled_dictio"]:
            self.data["compiled_dictio"].cleanup()

        if self.http_pool:
            self.http_pool.deregister()
            self.http_pool = None



import sys

# Python 2 and 3
if sys.version_info >= (3, 0):
    from urllib.parse import urljoin
else:
    from urlparse import urljoin

from wfuzz.plugin_api.mixins import DiscoveryPluginMixin
from wfuzz.plugin_api.base import BasePlugin
from wfuzz.plugin_api.urlutils import check_content_type
from wfuzz.externals.moduleman.plugin import moduleman_plugin


@moduleman_plugin
class robots(BasePlugin, DiscoveryPluginMixin):
    name = "robots"
    author = ("Xavi Mendez (@xmendez)",)
    version = "0.1"
    summary = "Parses robots.txt looking for new content."
    description = ("Parses robots.txt looking for new content.",)
    category = ["active", "discovery"]
    priority = 99

    parameters = ()

    def __init__(self):
        BasePlugin.__init__(self)

    def validate(self, fuzzresult):
        return (
            fuzzresult.history.urlparse.ffname == "robots.txt"
            and fuzzresult.code == 200
            and check_content_type(fuzzresult, "text")
        )

    def process(self, fuzzresult):
        # Shamelessly (partially) copied from w3af's plugins/discovery/robotsReader.py
        for line in fuzzresult.history.content.split("\n"):
            line = line.strip()

            if (
                len(line) > 0
                and line[0] != "#"
                and (
                    line.upper().find("ALLOW") == 0
                    or line.upper().find("DISALLOW") == 0
                    or line.upper().find("SITEMAP") == 0
                )
            ):

                url = line[line.find(":") + 1 :]
                url = url.strip(" *")

                if url:
                    new_link = urljoin(fuzzresult.url, url)
                    self.queue_url(new_link)
                    self.add_result("link", "New link found", new_link)



class FuzzException(Exception):
    pass


class FuzzExceptBadOptions(FuzzException):
    pass


class FuzzExceptNoPluginError(FuzzException):
    pass


class FuzzExceptPluginLoadError(FuzzException):
    pass


class FuzzExceptIncorrectFilter(FuzzException):
    pass


class FuzzExceptBadAPI(FuzzException):
    pass


class FuzzExceptInternalError(FuzzException):
    pass


class FuzzExceptBadFile(FuzzException):
    pass


class FuzzExceptBadInstall(FuzzException):
    pass


class FuzzExceptBadRecipe(FuzzException):
    pass


class FuzzExceptMissingAPIKey(FuzzException):
    pass


class FuzzExceptPluginBadParams(FuzzException):
    pass


class FuzzExceptResourceParseError(FuzzException):
    pass


class FuzzExceptPluginError(FuzzException):
    pass


class FuzzExceptNetError(FuzzException):
    pass



import re

from wfuzz.plugin_api.base import BasePlugin
from wfuzz.exception import FuzzExceptPluginBadParams
from wfuzz.externals.moduleman.plugin import moduleman_plugin


@moduleman_plugin
class npm_deps(BasePlugin):
    name = "npm_deps"
    author = ("Xavi Mendez (@xmendez)",)
    version = "0.1"
    summary = "Looks for npm dependencies definition in js code"
    description = (
        "Extracts npm packages by using regex pattern from the HTTP response and prints it",
    )
    category = ["info"]
    priority = 99

    parameters = ()

    REGEX_PATT = re.compile(r'"([^"]+)":"([^"]+)"', re.MULTILINE | re.DOTALL)
    REGEX_DEP = re.compile(
        r"dependencies:\{(.*?)\}", re.MULTILINE | re.DOTALL | re.IGNORECASE
    )
    REGEX_DEV_DEP = re.compile(
        r"devdependencies:\{(.*?)\}", re.MULTILINE | re.DOTALL | re.IGNORECASE
    )

    def __init__(self):
        BasePlugin.__init__(self)

    def validate(self, fuzzresult):
        if fuzzresult.history.urlparse.fext != ".js" or fuzzresult.code != 200:
            return False

        self.match = self.REGEX_DEP.search(fuzzresult.history.content)
        self.match_dev = self.REGEX_DEV_DEP.search(fuzzresult.history.content)

        return self.match is not None or self.match_dev is not None

    def process(self, fuzzresult):
        if self.match_dev:
            for name, version in self.REGEX_PATT.findall(self.match_dev.group(1)):
                self.add_result("dependency", "npm dependency", name)

        if self.match:
            for name, version in self.REGEX_PATT.findall(self.match.group(1)):
                self.add_result("dev_dependency", "npm dev dependency", name)
